/*********************************************************************
This file is part of the ccpro-1.0 software. 

Copyright (C) 2020  Pietro Di Lena

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include "mat.h"
#include "ccstat.h"

// GLOBAL OPTIONS
static int VERB = 0;		// Verbose on/off
static int TYPE = 0;		// 0 = emap, 1 = e2map, 3 = vmap
static int BFAL = 0;		// Brute force algorithm
static int RECM = 0;		// Reciprocal matrix
static unsigned int DIAG = 0;	// Remove diagonal elements < DIAG
static char *OUTF = NULL;	// Output file name
static double PREC = 0;		// Round to zero elements < PREC

static char *YMAP = NULL;
static char *MVAL = NULL;

static int read_ui(char *num, unsigned int *n) {
	char *ptr;
	int base = 10;

	*n = strtoul(num, &ptr, base);
	return num == ptr || errno != 0 || (errno == 0 && num && *ptr != 0);
}

static int read_d(char *num, double *n) {
	char *ptr;

	*n = strtod(num, &ptr);
	return num == ptr || errno != 0 || (errno == 0 && num && *ptr != 0);
}

void usage(void) {
	fprintf(stderr, "Usage: ccmap [options] <map file> <target size>\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "\t-t, --type=<t>       Computes map t=emap|e2map|vmap. Default: %s\n", TYPE == 0 ? "emap" : TYPE == 1 ? "e2map" : "vmap");
	fprintf(stderr, "\t-d, --diag=<n>       Ignore diagonals lower than <n> in the input matrix. Default: %u\n", DIAG);
	fprintf(stderr, "\t-o, --out=<file>     Save the matrix into <file>:. Default: %s\n", OUTF ? OUTF : "stdout");
	fprintf(stderr, "\t-p, --prec=<p>       Set to zero values lower than <p> in the output matrix. Default: %.0e\n", PREC);
	fprintf(stderr, "\t-r, --reciprocal     Use the reciprocal matrix. Default: %s\n", RECM == 0 ? "false" : "true");
	fprintf(stderr, "\t-b, --bf             Use brute force algorithm (for debugging only). Default: %s\n", BFAL ? "false" : "true");
	fprintf(stderr, "\t-v, --verbose        Verbose on. Default: %s\n", VERB ? "on" : "off");
	fprintf(stderr, "\t-h, --help           Print this help\n");
}

int parse_args(int argc, char *argv[]) {
	int opt = 0, long_index = 0;

	static struct option long_options[] = {
		{"type", required_argument, 0, 't'},
		{"diag", required_argument, 0, 'd'},
		{"out", required_argument, NULL, 'o'},
		{"prec", required_argument, 0, 'p'},
		{"reciprocal", no_argument, 0, 'r'},
		{"bf", no_argument, 0, 'b'},
		{"verbose", no_argument, 0, 'v'},
		{"help", no_argument, 0, 'h'},
		{0, 0, 0, 0}
	};

	while ((opt = getopt_long(argc, argv, "t:d:o:p:rbvh", long_options, &long_index)) != -1) {
		switch (opt) {
		case 't':
			if (strcmp(optarg, "emap") == 0) {
				TYPE = 0;
			} else if (strcmp(optarg, "e2map") == 0) {
				TYPE = 1;
			} else if (strcmp(optarg, "vmap") == 0) {
				TYPE = 2;
			} else {
				fprintf(stderr, "Unrecognized option %s\n", optarg);
				return 1;
			}
			break;
		case 'd':
			if (read_ui(optarg, &DIAG) != 0) {
				fprintf(stderr, "Error convertin -d argument: %s\n", optarg);
				return 1;
			}
			break;
		case 'o':
			OUTF = optarg;
			break;
		case 'p':
			if (read_d(optarg, &PREC) != 0) {
				fprintf(stderr, "Error converting -p argument: %s\n", optarg);
				return 1;
			}
			break;
		case 'r':
			RECM = 1;
			break;
		case 'b':
			BFAL = 1;
			break;
		case 'v':
			VERB = 1;
			break;
		case 'h':
			usage();
			return 0;
		default:
			fprintf(stderr, "Unrecognized option -%c\n\n", opt);
			return 1;
		}
	}

	if (argc - optind < 2) {
		usage();
		return 1;
	} else if (argc - optind > 2) {
		int i;
		fprintf(stderr, "Unrecognized arguments:");
		for (i = optind + 2; i < argc; i++)
			fprintf(stderr, " %s", argv[i]);
		fprintf(stderr, "\n\n");
		return 1;
	}

	YMAP = argv[optind];
	MVAL = argv[optind + 1];

	return 0;
}

int main(int argc, char *argv[]) {
	double **Y, **M;
	unsigned int m, n, N;

	if (parse_args(argc, argv) != 0)
		return 1;

	if ((Y = mat_read(YMAP, &n, DIAG)) == NULL) {
		fprintf(stderr, "Error reading %s\n", YMAP);
		return 1;
	}

	if (read_ui(MVAL, &m) != 0) {
		fprintf(stderr, "Error converting to unsigned int %s\n", MVAL);
		return 1;
	}

	if (RECM == 1)
		mat_reciprocal(Y, n);

	if (VERB)
		cc_verbose_on();
	if (TYPE == 0)
		M = BFAL ? cc_emap_bf(Y, m, n) : cc_emap(Y, m, n);
	else if (TYPE == 1)
		M = BFAL ? cc_e2map_bf(Y, m, n) : cc_e2map(Y, m, n);
	else
		M = BFAL ? cc_vmap_bf(Y, m, n) : cc_vmap(Y, m, n);
	if (VERB)
		cc_verbose_off();

	if (M == NULL) {
		fprintf(stderr, "Error: out of memory\n");
		return 1;
	}

	N = TYPE == 0 ? m : m * m;

	if (PREC != 0)
		mat_round(M, N, PREC);

	if (OUTF != NULL)
		mat_save(M, N, OUTF);
	else
		mat_full_print(M, N);

	mat_double_free(M);

	return 0;
}
