/*********************************************************************
This file is part of the ccpro-1.0 software. 

Copyright (C) 2020  Pietro Di Lena

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include "mat.h"
#include "ccstat.h"

#define N 10			// buffer size for printing

static int VERB = 0;		// Verbose on/off
static int RECM = 0;		// Reciprocal matrix
static int CC = 0;		// Print congruence coefficient
static int MEAN = 0;		// Print mean
static int VAR = 0;		// Print var
static int PERM = 0;		// Print permutation mean and var
static int SAMP = 0;		// Print sampling mean and var
static unsigned int NSAM = 10000;	// Number of sampling
static int BFAL = 0;		// Brute-force computation
static unsigned int DIAG = 0;
static char *ALN = NULL;
static char *EMAP = NULL;
static char *E2MAP = NULL;
static char *VMAP = NULL;
static char *XMAP = NULL;
static char *YMAP = NULL;

static int read_ui(char *num, unsigned int *n) {
	char *ptr;
	int base = 10;

	*n = strtoul(num, &ptr, base);
	return num == ptr || errno != 0 || (errno == 0 && num && *ptr != 0);
}

void usage_full(void) {
	fprintf(stderr, "Usage: ccpro [options] <map1 file> <map2 file>\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "\n**** Output options ***\n");
	fprintf(stderr, "\t-C, --print_cc      Print congruence coefficient (needs an alignment or map1 and map2 with same size).\n\n");
	fprintf(stderr, "\t                    Output. cc: <conguence coefficient> P-val1: <p-value map1> P-val2: <p-value map2>\n\n");
	fprintf(stderr, "\t                    P-val1 gives the probability of uniformly samply a random symmetric matrix map3, of\n");
	fprintf(stderr, "\t                    same size as map2, such that cc(map1,map3) > cc(map1,map2). P-val2 is the same for map2.\n\n");
	fprintf(stderr, "\t                    Input maps can be either in CASP format or simple format. Simple format:\n");
	fprintf(stderr, "\t                          <map size>\n");
	fprintf(stderr, "\t                          <i> <j> <value>\n");
	fprintf(stderr, "\t                          ...\n");
	fprintf(stderr, "\t                    where 0 <= i <= j < map size\n\n");
	fprintf(stderr, "\t-M, --print_mean    Print mean over all alignments. Default: %s\n\n", MEAN == 0 ? "FALSE" : "TRUE");
	fprintf(stderr, "\t                    Output. E[cc]: <mean congruence coefficient over all alignments>\n\n");
	fprintf(stderr, "\t-D, --print_sd      Print standard deviation over all alignments. Default: %s\n\n", VAR == 0 ? "FALSE" : "TRUE");
	fprintf(stderr, "\t                    Output. Sd[cc]: <standard deviation over all alignments>\n\n");
	fprintf(stderr, "\t                    If -CMD options produce output the alignment Z-score: (cc-E[cc])/Sd[cc] is also shown.\n\n");
	fprintf(stderr, "\t-P, --print_perm    Print mean and standard deviation over all permutations. Default: %s\n\n", PERM == 0 ? "FALSE" : "TRUE");
	fprintf(stderr, "\t                    Output. EP[cc]: <mean over permutations> SdP[cc]: <standard deviation over permutations>\n\n");
	fprintf(stderr, "\t                    If -CP options produce output the alignment Z-scoreP: (cc-EP[cc])/SdP[cc] is also shown.\n\n");
	fprintf(stderr, "\t-S, --print_samp    Print sampling mean and standard deviation. Default: %s\n\n", SAMP == 0 ? "FALSE" : "TRUE");
	fprintf(stderr, "\t                    Output. ES[cc]: <sampling mean> SdS[cc]: <sampling standard deviation>\n\n");
	fprintf(stderr, "\t                    If -CS options produce output the alignment Z-scoreS: (cc-ES[cc])/SdS[cc] is also shown.\n\n");
	fprintf(stderr, "\t-v, --verbose       Verbose on. Default: %s\n\n", VERB ? "on" : "off");
	fprintf(stderr, "\t                    Useful for -MDSb options to see the ETA.\n\n");
	fprintf(stderr, "\t-h, --help          Print short help and exit\n");
	fprintf(stderr, "\t-H, --fullhelp      Print this help and exit\n");
	fprintf(stderr, "\n**** Input options ***\n");
	fprintf(stderr, "\t-A, --aln=<aln>     Alignment between map1 and map2.\n\n");
	fprintf(stderr, "\t                    Alignment can be either in fasta format or simple format. Simple format:\n");
	fprintf(stderr, "\t                          <map1 size> <map2 size>\n");
	fprintf(stderr, "\t                          <i1>:<j1>\n");
	fprintf(stderr, "\t                          ...\n");
	fprintf(stderr, "\t                    where 0 <= i1 < map1 size, 0 <= i2 < map2 size. Alignment consistency is checked.\n\n");
	fprintf(stderr, "\t-N, --nsamp=<n>     Set number of sampled alignments. Default: %u\n\n", NSAM);
	fprintf(stderr, "\t                    If n is larger than #alignments, then n is set to #alignments.\n\n");
	fprintf(stderr, "\t-E, --emap=<map>    Expectation matrix of map2.\n\n");
	fprintf(stderr, "\t                    Precomputed expectation matrix E[map2] of size m x m, where m is the size of map1.\n");
	fprintf(stderr, "\t                    Used to speed up -MD options.\n\n");
	fprintf(stderr, "\t-2, --e2map=<map>   Expectation2 matrix of map.c\n\n");
	fprintf(stderr, "\t                    Precomputed expectation mtrix E[map2 * map2] of size m^2 x m^2, where m is the size of map1\n");
	fprintf(stderr, "\t                    and * is the Kronecker product. Used to speed up -D option.\n\n");
	fprintf(stderr, "\t-V, --vmap=<map>    Variance-covariance matrix of map2.\n\n");
	fprintf(stderr, "\t                    Precomputed variance-covariance matrix Var[map2] = E[map2 * map2] - E[map2]*E[map2] of size m^2 x m^2\n");
	fprintf(stderr, "\t                    where m is the size of map1 and and * is the Kronecker product. Used to speed up -D option.\n\n");
	fprintf(stderr, "\t-d, --diag=<n>      Ignore diagonals lower than <n> in the input matrix. Default: %u\n\n", DIAG);
	fprintf(stderr, "\t                    Diagonal 0 is the main diagonal. Entries of diagonals lower than <n> are set to 0.\n\n");
	fprintf(stderr, "\t-r, --reciprocal    Use the reciprocal matrix. Default: %s\n\n", RECM ? "true" : "false");
	fprintf(stderr, "\t                    In reciprocal matrix every non-zero entry x is recoded as 1/x.\n\n");
	fprintf(stderr, "\t-b, --bf            Use brute-force algorithm for -M -V -P (debugging only). Default: %s\n\n", BFAL ? "true" : "false");
	fprintf(stderr, "\t                    Brute force algorithm evaluates all alignments (-MV) or all permutations (-P). Exponential time\n");
	fprintf(stderr, "\t                    algorithm, feasible only on small maps.\n\n");
}

void usage(void) {
	fprintf(stderr, "Usage: ccpro [options] <map1 file> <map2 file>\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "\n**** Output options ***\n");
	fprintf(stderr, "\t-C, --print_cc      Print congruence coefficient (needs an alignment or map1 and map2 with same size).\n");
	fprintf(stderr, "\t-M, --print_mean    Print mean over all alignments. Default: %s\n", MEAN == 0 ? "FALSE" : "TRUE");
	fprintf(stderr, "\t-D, --print_sd      Print standard deviation over all alignments. Default: %s\n", VAR == 0 ? "FALSE" : "TRUE");
	fprintf(stderr, "\t-P, --print_perm    Print mean and standard deviation over all permutations. Default: %s\n", PERM == 0 ? "FALSE" : "TRUE");
	fprintf(stderr, "\t-S, --print_samp    Print sampling mean and standard deviation. Default: %s\n", SAMP == 0 ? "FALSE" : "TRUE");
	fprintf(stderr, "\t-v, --verbose       Verbose on. Default: %s\n", VERB ? "on" : "off");
	fprintf(stderr, "\t-h, --help          Print this help and exit\n");
	fprintf(stderr, "\t-H, --fullhelp      Print full help and exit\n");
	fprintf(stderr, "\n**** Input options ***\n");
	fprintf(stderr, "\t-A, --aln=<aln>     Alignment between map1 and map2\n");
	fprintf(stderr, "\t-N, --nsamp=<n>     Set number of sampled alignments. Default: %u\n", NSAM);
	fprintf(stderr, "\t-E, --emap=<map>    Expectation matrix of map2\n");
	fprintf(stderr, "\t-2, --e2map=<map>   Expectation2 matrix of map2\n");
	fprintf(stderr, "\t-V, --vmap=<map>    Variance-covariance matrix of map2\n");
	fprintf(stderr, "\t-d, --diag=<n>      Ignore diagonals lower than <n> in the input matrix. Default: %u\n", DIAG);
	fprintf(stderr, "\t-r, --reciprocal    Use the reciprocal matrix. Default: %s\n", RECM ? "true" : "false");
	fprintf(stderr, "\t-b, --bf            Use brute-force algorithm for -M -V -P (debugging only). Default: %s\n", BFAL ? "true" : "false");
}

int parse_args(int argc, char *argv[]) {
	int opt = 0, long_index = 0;

	static struct option long_options[] = {
		{"aln", required_argument, 0, 'A'},
		{"print_cc", no_argument, 0, 'C'},
		{"print_mean", no_argument, 0, 'M'},
		{"print_sd", no_argument, 0, 'D'},
		{"print_perm", no_argument, 0, 'P'},
		{"print_samp", no_argument, 0, 'S'},
		{"nsamp", required_argument, 0, 'N'},
		{"emap", required_argument, NULL, 'E'},
		{"e2map", required_argument, NULL, '2'},
		{"vmap", required_argument, NULL, 'V'},
		{"diag", required_argument, 0, 'd'},
		{"reciprocal", no_argument, 0, 'r'},
		{"bf", no_argument, 0, 'b'},
		{"verbose", no_argument, 0, 'v'},
		{"help", no_argument, 0, 'h'},
		{"fullhelp", no_argument, 0, 'H'},
		{0, 0, 0, 0}
	};

	while ((opt = getopt_long(argc, argv, "A:CMDPSN:E:2:V:d:brvhH", long_options, &long_index)) != -1) {
		switch (opt) {
		case 'A':
			ALN = optarg;
			break;
		case 'C':
			CC = 1;
			break;
		case 'M':
			MEAN = 1;
			break;
		case 'D':
			VAR = 1;
			break;
		case 'P':
			PERM = 1;
			break;
		case 'S':
			SAMP = 1;
			break;
		case 'N':
			if (read_ui(optarg, &NSAM) != 0) {
				fprintf(stderr, "Error reading -n argument: %s\n", optarg);
				return 1;
			}
			break;
		case 'E':
			EMAP = optarg;
			break;
		case '2':
			E2MAP = optarg;
			break;
		case 'V':
			VMAP = optarg;
			break;
		case 'd':
			if (read_ui(optarg, &DIAG) != 0) {
				fprintf(stderr, "Error reading -d argument: %s\n", optarg);
				return 1;
			}
			break;
		case 'b':
			BFAL = 1;
			break;
		case 'r':
			RECM = 1;
			break;
		case 'v':
			VERB = 1;
			break;
		case 'h':
			usage();
			exit(0);
		case 'H':
			usage_full();
			exit(0);
		default:
			fprintf(stderr, "Unrecognized option -%c\n\n", opt);
			return 1;
		}
	}

	if (argc - optind < 2) {
		usage();
		return 1;
	} else if (argc - optind > 2) {
		int i;
		fprintf(stderr, "Unrecognized arguments:");
		for (i = optind + 2; i < argc; i++)
			fprintf(stderr, " %s", argv[i]);
		fprintf(stderr, "\n\n");
		return 1;
	}
	XMAP = argv[optind];
	YMAP = argv[optind + 1];

	return 0;
}

int main(int argc, char *argv[]) {
	double **X, **Y;
	unsigned int m, n;
	int *A = NULL;		// Alignment
	double **E = NULL, **E2 = NULL, **V = NULL;
	double ccscore = NAN, ccpval1 = NAN, ccpval2 = NAN;
	double mean = NAN, var = NAN;
	double pmean = NAN, pvar = NAN;
	double smean = NAN, svar = NAN;

	if (parse_args(argc, argv) != 0)
		return 1;

	if ((X = mat_read(XMAP, &m, DIAG)) == NULL) {
		fprintf(stderr, "Error reading %s\n", XMAP);
		return 1;
	}

	if ((Y = mat_read(YMAP, &n, DIAG)) == NULL) {
		fprintf(stderr, "Error reading %s\n", YMAP);
		return 1;
	}

	if (RECM == 1) {
		mat_reciprocal(X, m);
		mat_reciprocal(Y, n);
	}
	// Read the alignment
	if (ALN != NULL) {
		unsigned int mtmp, ntmp;
		if ((A = mat_aln_read(ALN, &mtmp, &ntmp)) == NULL) {
			fprintf(stderr, "Error: alignment %s not readable or inconsistent\n", ALN);
			return 1;
		}
		if (mtmp != m || ntmp != n) {
			fprintf(stderr, "Error: alignment %s not consistent with input maps\n", ALN);
			return 1;
		}
	}
	// Read the expectation matrix
	if ((MEAN || VAR) && EMAP != NULL) {
		unsigned int tmp;
		if ((E = mat_read(EMAP, &tmp, 0)) == NULL) {
			fprintf(stderr, "Error reading %s\n", EMAP);
			return 1;
		}
		if (m != tmp) {
			fprintf(stderr, "Error: expectation matrix (%s) size is not compatible with %s\n", EMAP, XMAP);
			return 1;
		}
	}
	// Read the expectation2 matrix
	if (VAR && E2MAP != NULL) {
		unsigned int tmp;
		if ((E2 = mat_read(E2MAP, &tmp, 0)) == NULL) {
			fprintf(stderr, "Error reading %s\n", E2MAP);
			return 1;
		}
		if (m * m != tmp) {
			fprintf(stderr, "Error: expectation2 matrix (%s) size is not compatible with %s\n", E2MAP, XMAP);
			return 1;
		}
	}
	// Read the variance-covariance matrix
	if (VAR && VMAP != NULL) {
		unsigned int tmp;
		if ((V = mat_read(VMAP, &tmp, 0)) == NULL) {
			fprintf(stderr, "Error reading %s\n", VMAP);
			return 1;
		}
		if (m * m != tmp) {
			fprintf(stderr, "Error: variance-covariance matrix (%s) size is not compatible with %s\n", VMAP, XMAP);
			return 1;
		}
	}
	// Compute the congruence coefficient
	if (CC) {
		if (A != NULL) {
			ccscore = cc_aln(X, Y, m, n, A);
		} else if (m == n) {
			ccscore = cc(X, Y, m);
		} else if (VERB) {
			fprintf(stderr, "Warning: cannont compute the congruence coefficient since map1 and map2\n");
			fprintf(stderr, "         have different sizes and no alignment has been provided\n");
		}

	}

	if (VERB)
		cc_verbose_on();

	// Compute the mean over all alignments
	if (MEAN) {
		if (BFAL) {
			if (VERB && (m > 20 || n > 20))
				fprintf(stderr, "Warning: brute force algorithm for large matrices is too slow!\n");
			cc_mv_bf(X, Y, m, n, &mean, &var);
		} else {
			mean = cc_mean(X, Y, m, n, E);
		}
	}
	// Compute the variance over all alignments
	if (VAR) {
		if (BFAL && isnan(var)) {
			if (VERB && (m > 20 || n > 20))
				fprintf(stderr, "Warning: brute force algorithm for large matrices is too slow!\n");
			cc_mv_bf(X, Y, m, n, &mean, &var);
		} else {
			var = cc_var(X, Y, m, n, E, E2, V);
		}
	}
	// Compute mean and variance over all permutations
	if (PERM) {
		if (BFAL) {
			if (VERB && m + n > 10)
				fprintf(stderr, "Warning: brute force algorithm for large matrices is too slow!\n");
			cc_mv_perm_bf(X, Y, m, n, &pmean, &pvar);
		} else {
			pmean = cc_mean_perm(X, Y, m, n);
			pvar = cc_var_perm(X, Y, m, n);
		}
	}
	// Compute sample mean and variance
	if (SAMP)
		cc_mv_samp(X, Y, m, n, NSAM, &smean, &svar);

	if (VERB)
		cc_verbose_off();

	if (!isnan(ccscore)) {
		unsigned int df1 = mat_nonzeros(X, m);
		unsigned int df2 = mat_nonzeros(Y, n);
		ccpval1 = cc_pval(ccscore, df1);
		ccpval2 = cc_pval(ccscore, df2);
	}

	if (!isnan(ccscore)) {
		char buf[N];
		printf("cc: %.4f ", ccscore);
		sprintf(buf, "%2.2e", ccpval1);
		printf("P-val1: %-*s ", N - 1, buf);
		sprintf(buf, "%2.2e", ccpval2);
		printf("P-val2: %-*s ", N - 1, buf);
	}

	if (!isnan(mean))
		printf("E[cc]: %.4f ", mean);
	if (!isnan(var))
		printf("Sd[cc]: %.4f ", sqrt(var));
	if (!isnan(ccscore) && !isnan(mean) && !isnan(var))
		printf("Z-score: %2.4f ", var == 0 ? NAN : (ccscore - mean) / sqrt(var));

	if (!isnan(smean))
		printf("ES[cc]: %.4f ", smean);
	if (!isnan(svar))
		printf("SdS[cc]: %.4f ", sqrt(svar));
	if (!isnan(ccscore) && !isnan(smean) && !isnan(svar))
		printf("Z-scoreS: %2.4f ", svar == 0 ? NAN : (ccscore - smean) / sqrt(svar));

	if (!isnan(pmean))
		printf("EP[cc]: %.4f ", pmean);
	if (!isnan(pvar))
		printf("SdP[cc]: %.4f ", sqrt(pvar));
	if (!isnan(ccscore) && !isnan(pmean) && !isnan(pvar))
		printf("Z-scoreP: %2.4f ", pvar == 0 ? NAN : (ccscore - pmean) / sqrt(pvar));
	printf("\n");

	return 0;
}
