/*********************************************************************
This file is part of the ccpro-1.0 software. 

Copyright (C) 2020  Pietro Di Lena

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
**********************************************************************/

#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_errno.h>
#include "ccstat.h"
#include "mat.h"

#define MAX(x,y) ((x)<(y)?(y):(x))
#define MIN(x,y) ((x)>(y)?(y):(x))
#define SAME_DIST(v,p,n) (v[0]==p[0] && (n>1?v[1]==p[1]:1) && (n>2?v[2]==p[2]:1) && (n>3?v[3]==p[3]:1))
#define SWAP(P,i,j) {int tmp = P[i]; P[i] = P[j]; P[j] = tmp;}

static long double **LOGBIN = NULL;

static unsigned int VERB = 0;

/**************************************
        PRIVATE FUNCTIONS PROTOTYPES
 **************************************/

// Setup expectation matrix E[cc^2(X,Y)] for 1 match
static void e2map_match1(struct index *I, int start, int end, int m, int n, double **E2);

// Setup expectation matrix E[cc^2(X,Y)] for 2 matches
static void e2map_match2(struct index *I, int start, int end, int m, int n, double **E2);

// Setup expectation matrix E[cc^2(X,Y)] for 3 matches
static void e2map_match3(struct index *I, int start, int end, int m, int n, double **E2);

// Setup expectation matrix E[cc^2(X,Y)] for 4 matches
static void e2map_match4(struct index *I, int start, int end, int m, int n, double **E2);

// Mean congruence coefficient over all permutations for symmetric matrices with zero main diagonal.
// N should be >= max(m,n). If N=m+n permutations include all allignments alpha : {1,..,m} -> {1,..,n}
static double cc_mean_perm_zero_diag(double **X, double **Y, unsigned int m, unsigned int n, unsigned int N);

// Mean congruence coefficient over all permutations for general symmetric matrices.
// N should be >= max(m,n). If N=m+n permutations include all allignments alpha : {1,..,m} -> {1,..,n}
static double cc_mean_perm_nonzero_diag(double **X, double **Y, unsigned int m, unsigned int n, unsigned int N);

// Variance of the congruence coefficient over all permutations for symmetric matrices with zero main diagonal.
// N should be >= max(m,n). If N=m+n permutations include all allignments alpha : {1,..,m} -> {1,..,n}
static double cc_var_perm_zero_diag(double **X, double **Y, unsigned int m, unsigned int n, unsigned int N);

// Variance of the congruence coefficient over all permutations for general symmetric matrices.
// N should be >= max(m,n). If N=m+n permutations include all allignments alpha : {1,..,m} -> {1,..,n}
static double cc_var_perm_nonzero_diag(double **X, double **Y, unsigned int m, unsigned int n, unsigned int N);

// Sample num random alignments alpha : {1,..,m} -> {1,..,n} of size k (i.e. k matches)
static int **cc_mv_samp_alignments(unsigned int m, unsigned int n, unsigned int k, unsigned int num);

// Free algnment matrix A
static void cc_mv_samp_free_alignments(int **A, unsigned int num);

// Accumulate cc and cc^2 coefficients with respect to alignments A into tmpE and tmpV, respectively
static void cc_mv_samp_add(double **X, double **Y, unsigned int m, int **A, unsigned int num, double *tmpE, double *tmpV);

// Log frequency of k-length alignments for k=min(m,(m,n)
static long double *cc_mv_samp_len_distr(unsigned int m, unsigned int n);

// Log binomial coefficient m choose k
static long double log_choose(unsigned int m, unsigned int k);

// Log(n!)
static long double log_fact(unsigned int n);

// Sample a random alignment alpha : {1,..,m} -> {1,..,n} of size k (i.e. k matches)
static int *random_alignment(unsigned int m, unsigned int n, unsigned int k);

// Sample k random positions without repetition in {1,..,m} 
static int *random_positions(unsigned int m, unsigned int k);

// Returns 1 if alignments A and B are equal, 0 otherwise
static int same_alignments(int *A, int *B, unsigned int m);

// Setup matrix of log number of alignments for lengths 0<=i<=m and 0<=j<=n 
static void logbin_matrix_setup(unsigned int m, unsigned int n);

// Free logbin matrix
static void logbin_matrix_free(void);

// Counts the fraction of alignments for given matched positions.
// - p1 and p2: odered mapped positions (starting from 0)
// - n: number of mapped positions+1, i.e. length of p1 and p2
// - p1[n-1]: length of the first sequence
// - p2[n-1]: length of the second sequence 
static long double numaln_fraction(unsigned int p1[], unsigned int p2[], unsigned int n);

// Counts the fraction of alignments for one matched position.
// - i, k are the matched positions
// - 0 <= i < m
// - 0 <= j < n
static long double numaln_fraction_match1(unsigned int i, unsigned int k, unsigned int m, unsigned int n);

// Counts the fraction of alignments for two matched positions.
// - i, k are the first matched positions
// - j, l are the second matched positions
// - 0 <= i,k < m
// - 0 <= j,l < n
static long double numaln_fraction_match2(unsigned int i, unsigned int j, unsigned int k, unsigned int l, unsigned int m, unsigned int n);

// Heap algorithm for generatig all permutations
static void heap_perm(double **X, double **Y, unsigned int m, unsigned int n, int *P, unsigned int k, double *tot_perm, double *tot_cc, double *tot_cc2);

// Brute force recursive procedure for expectation matrix calculation
static void cc_emap_bf_rec(double **Y, unsigned int m, unsigned int n, int *A, unsigned int mi, unsigned int ni, double *tot_aln, double **E);

// Brute force recursive procedure for expectation2 matrix calculation
void cc_e2map_bf_rec(double **Y, unsigned int m, unsigned int n, int *A, unsigned int mi, unsigned int ni, double *tot_aln, double **E2);

// Brute force recursive procedure for mean and variance calculation
static void cc_mv_bf_rec(double **X, double **Y, unsigned int m, unsigned int n, int *A, unsigned int mi, unsigned int ni, double *tot_aln, double *tot_cc, double *tot_cc2);

// Estimate ETA from current stats
inline static char *ETA(clock_t start_t, clock_t end_t, unsigned int count, unsigned int size);

// Print ETA for matrix computation. Needs verbose on
static void print_eta(unsigned int N, unsigned int count, unsigned int size);

/******************************
	PUBLIC FUNCTIONS
 ******************************/

double **cc_emap(double **Y, unsigned int m, unsigned int n) {
	unsigned int i, j, k = 0, indx_len;
	double **L = mat_to_list(Y, n, &indx_len);
	double **E, tr = sqrt(mat_trace_prod(Y, Y, n));

	logbin_matrix_setup(m, n);

	if ((E = mat_double_alloc(m, m)) == NULL)
		return NULL;	// Failed allocation

	if (tr == 0) {
		if (VERB)
			fprintf(stderr, "WARNING: input matrix is zero\n");
	} else {
		if (VERB)
			fprintf(stderr, "*** SETTING UP EMAP ***\n\n#Alignments: %LE\n\n", expl(LOGBIN[m][n]));
		if (VERB)
			print_eta(0, 0, indx_len);
		for (k = 0; k < indx_len; k++) {
			if (L[k][0] == L[k][1]) {
				for (i = 0; i < m; i++)
					E[i][i] += numaln_fraction_match1(i, L[k][0], m, n) * L[k][2];
			} else {
				for (i = 0; i < m; i++)
					for (j = i + 1; j < m; j++)
						E[i][j] += numaln_fraction_match2(i, j, L[k][0], L[k][1], m, n) * L[k][2];
			}
			if (VERB)
				print_eta(0, k + 1, indx_len);
		}
		if (VERB)
			fprintf(stderr, "\n");

		for (i = 0; i < m; i++)
			for (j = i; j < m; j++)
				E[i][j] = E[j][i] = E[i][j] / tr;

	}

	logbin_matrix_free();
	mat_double_free(L);

	return E;
}

double **cc_e2map(double **Y, unsigned int m, unsigned int n) {
	struct index *I;
	unsigned int i, j, indx_len, start_pos[] = { -1, -1, -1, -1, -1 }, end_pos[] = {
	-1, -1, -1, -1, -1};
	double **E2, tr = mat_trace_prod(Y, Y, n);

	logbin_matrix_setup(m, n);

	I = mat_index(Y, n, &indx_len);

	for (i = 0; i < indx_len; i++) {
		int pos = I[i].n;
		if (start_pos[pos] == -1)
			start_pos[pos] = i;
		end_pos[pos] = i;
	}

	if ((E2 = mat_double_alloc(m * m, m * m)) == NULL)
		return NULL;	// Failed allocation

	if (VERB)
		fprintf(stderr, "*** SETTING UP E2MAP ***\n\n#Alignments: %LE\n\n", expl(LOGBIN[m][n]));
	if (start_pos[1] != -1)
		e2map_match1(I, start_pos[1], end_pos[1], m, n, E2);
	if (start_pos[2] != -1)
		e2map_match2(I, start_pos[2], end_pos[2], m, n, E2);
	if (start_pos[3] != -1)
		e2map_match3(I, start_pos[3], end_pos[3], m, n, E2);
	if (start_pos[4] != -1)
		e2map_match4(I, start_pos[4], end_pos[4], m, n, E2);
	if (VERB)
		fprintf(stderr, "\n");

	for (i = 0; i < m * m; i++)
		for (j = i; j < m * m; j++)
			E2[i][j] = E2[j][i] = E2[i][j] / tr;

	logbin_matrix_free();
	free(I);

	return E2;
}

double **cc_vmap(double **Y, unsigned int m, unsigned int n) {
	double **E = cc_emap(Y, m, n);
	double **E2 = cc_e2map(Y, m, n);
	double **EE = mat_kronecker_prod(E, E, m);

	mat_sub(E2, EE, m * m);
	mat_double_free(E);
	mat_double_free(EE);

	return E2;
}

double cc_mean(double **X, double **Y, unsigned int m, unsigned int n, double **E) {
	double trXX = mat_trace_prod(X, X, m);
	double **Etmp = E == NULL ? cc_emap(Y, m, n) : E;
	double trXE = mat_trace_prod(X, Etmp, m);

	if (E == NULL)
		mat_double_free(Etmp);

	return trXE / sqrt(trXX);
}

double cc_mean_perm(double **X, double **Y, unsigned int m, unsigned int n) {
	if (mat_zero_diag(X, m) && mat_zero_diag(Y, n))
		return cc_mean_perm_zero_diag(X, Y, m, n, m + n);
	else
		return cc_mean_perm_nonzero_diag(X, Y, m, n, m + n);
}

double cc_var(double **X, double **Y, unsigned int m, unsigned int n, double **E, double **E2, double **V) {
	double **X2 = mat_kronecker_prod(X, X, m);
	double trXX = mat_trace_prod(X, X, m);
	double var;

	if (V != NULL) {
		var = mat_trace_prod(X2, V, m * m) / trXX;
	} else {
		double mean = cc_mean(X, Y, m, n, E);
		double **E2tmp = E2 == NULL ? cc_e2map(Y, m, n) : E2;

		var = mat_trace_prod(X2, E2tmp, m * m) / trXX - mean * mean;

		if (E2 == NULL)
			mat_double_free(E2tmp);
	}

	return var;
}

double cc_var_perm(double **X, double **Y, unsigned int m, unsigned int n) {
	if (mat_zero_diag(X, m) && mat_zero_diag(Y, n))
		return cc_var_perm_zero_diag(X, Y, m, n, m + n);
	else
		return cc_var_perm_nonzero_diag(X, Y, m, n, m + n);
}

unsigned int cc_mv_samp(double **X, double **Y, unsigned int m, unsigned int n, unsigned int samp_size, double *mean, double *var) {
	double trXX = mat_trace_prod(X, X, m);
	double trYY = mat_trace_prod(Y, Y, n);
	unsigned int tot = 0;

	srandom(time(NULL));

	*mean = *var = 0;

	if (trXX * trYY != 0) {
		unsigned int k, num;
		long double *P = cc_mv_samp_len_distr(m, n), tot_aln = expl(log_choose(m + n, MIN(m, n)));
		double Etmp = 0, Vtmp = 0;
		int **A;

		if (VERB)
			fprintf(stderr, "*** SETTING UP MEAN AND VAR WITH SAMPLING ALGORITHM ***\n\n#Alignments: %LE #Samplings: %u\n\n", tot_aln, samp_size);

		if (samp_size > tot_aln) {
			if (VERB)
				fprintf(stderr, "WARNING: #samp=%u > #align=%Lg. Recoding #samp = #align\n\n", samp_size, tot_aln);
			samp_size = round(tot_aln);
		}

		for (k = 0; k <= MIN(m, n); k++) {
			tot += num = samp_size == round(tot_aln) ? ceil(P[k] * samp_size) : round(P[k] * samp_size);

			if (VERB)
				fprintf(stderr, "- Sampling %u (%.2Lg%%) alignments of length %u\n", num, 100 * P[k], k);
			A = cc_mv_samp_alignments(m, n, k, num);
			cc_mv_samp_add(X, Y, m, A, num, &Etmp, &Vtmp);
			cc_mv_samp_free_alignments(A, num);
		}
		if (VERB)
			fprintf(stderr, "\n");

		Etmp /= tot;
		Vtmp /= (tot == samp_size ? tot : tot - 1);	// Exact or sample variance
		*mean = Etmp / sqrt(trXX * trYY);
		*var = (Vtmp - Etmp * Etmp) / (trXX * trYY);
	}

	return tot;
}

double cc(double **X, double **Y, unsigned int m) {
	double trXX = mat_trace_prod(X, X, m);
	double trYY = mat_trace_prod(Y, Y, m);
	double trXY = mat_trace_prod(X, Y, m);
	double norm = sqrt(trXX * trYY);

	return norm == 0 ? 0 : trXY / norm;
}

double cc_gen(double **X, double **Y, unsigned int m, unsigned int n) {
	double trXX = mat_trace_prod_gen(X, X, m, n);
	double trYY = mat_trace_prod_gen(Y, Y, m, n);
	double trXY = mat_trace_prod_gen(X, Y, m, n);
	double norm = sqrt(trXX * trYY);

	return norm == 0 ? 0 : trXY / norm;
}

double cc_aln(double **X, double **Y, unsigned int m, unsigned int n, int *A) {
	double trXX = mat_trace_prod(X, X, m);
	double trYY = mat_trace_prod(Y, Y, n);
	double trXY = mat_trace_prod_aln(X, Y, A, m);
	double norm = sqrt(trXX * trYY);

	return norm == 0 ? 0 : trXY / norm;
}

double cc_pval(double cc, unsigned int df) {
	// prevents exit on underflow
	gsl_set_error_handler_off();
	if (df == 0) {
		return 1;
	} else {
		double p = 0.5 * gsl_sf_beta_inc((df + 1) / 2.0, 0.5, 1 - cc * cc);
		return cc < 0 ? 1 - p : p;
	}
}

double cc_pval_gen(double cc, unsigned int df) {
	// prevents exit on underflow
	gsl_set_error_handler_off();
	if (df == 0) {
		return 1;
	} else {
		double p = 0.5 * gsl_sf_beta_inc((df - 1) / 2.0, 0.5, 1 - cc * cc);
		return cc < 0 ? 1 - p : p;
	}
}

void cc_verbose_on(void) {
	VERB = 1;
}

void cc_verbose_off(void) {
	VERB = 0;
}

/*****************************
	TESTING FUNCTIONS
 *****************************/

// Brute force algorithms

double **cc_emap_bf(double **Y, unsigned m, unsigned n) {
	int *A = (int *)calloc(m, sizeof(int));
	double **E = mat_double_alloc(m, m), tr = sqrt(mat_trace_prod(Y, Y, n)), tot_aln = 0;
	unsigned int i, j;

	cc_emap_bf_rec(Y, m, n, A, 0, 0, &tot_aln, E);

	for (i = 0; i < m; i++)
		for (j = i; j < m; j++)
			E[i][j] = E[j][i] = E[i][j] / (tot_aln * tr);

	free(A);
	return E;
}

double **cc_e2map_bf(double **Y, unsigned m, unsigned n) {
	int *A = (int *)calloc(m, sizeof(int));
	double **E2 = mat_double_alloc(m * m, m * m), tr = mat_trace_prod(Y, Y, n), tot_aln = 0;
	unsigned int i, j;

	cc_e2map_bf_rec(Y, m, n, A, 0, 0, &tot_aln, E2);

	for (i = 0; i < m * m; i++)
		for (j = i; j < m * m; j++)
			E2[i][j] = E2[j][i] = E2[i][j] / (tot_aln * tr);

	free(A);
	return E2;
}

double **cc_vmap_bf(double **Y, unsigned m, unsigned n) {
	double **E = cc_emap_bf(Y, m, n);
	double **E2 = cc_e2map_bf(Y, m, n);
	double **EE = mat_kronecker_prod(E, E, m);

	mat_sub(E2, EE, m * m);

	mat_double_free(E);
	mat_double_free(EE);
	return E2;
}

void cc_mv_bf(double **X, double **Y, unsigned int m, unsigned int n, double *mean, double *var) {
	int *A = (int *)calloc(m, sizeof(int));
	double tot_cc = 0, tot_cc2 = 0, tot_aln = 0;

	cc_mv_bf_rec(X, Y, m, n, A, 0, 0, &tot_aln, &tot_cc, &tot_cc2);

	*mean = tot_cc / (tot_aln);
	*var = tot_cc2 / (tot_aln) - (*mean * *mean);

	free(A);
}

void cc_mv_perm_bf(double **X, double **Y, unsigned int m, unsigned int n, double *mean, double *var) {
	int i, *P = (int *)calloc(m + n, sizeof(int));
	double tot_cc = 0, tot_cc2 = 0;
	double **YE = mat_extend(Y, n, m + n);
	double tot_perm = 0;

	for (i = 0; i < m + n; i++)
		P[i] = i;

	heap_perm(X, YE, m, n, P, m + n, &tot_perm, &tot_cc, &tot_cc2);

	*mean = tot_cc / (tot_perm);
	*var = tot_cc2 / (tot_perm) - (*mean * *mean);

	mat_double_free(YE);
	free(P);
}

/*****************************
        SUPPORT FUNCTIONS
 *****************************/

static void e2map_match1(struct index *I, int start, int end, int m, int n, double **E2) {
	unsigned int N = 1;
	if (m >= N && n >= N) {
		unsigned int a, size = round(expl(log_choose(m, N))), count = 0;

		if (VERB)
			print_eta(N, 0, size);
		for (a = 0; a < m; a++) {
			int i = start;
			unsigned int p1[] = { a, m };
			while (i <= end) {
				unsigned int p2[5] = { I[i].val[0], n };
				long double num_aln = numaln_fraction(p1, p2, N + 1);

				while (i <= end && SAME_DIST(I[i].val, p2, I[i].n)) {
					int x = p1[I[i].ind[0]] * m + p1[I[i].ind[2]];
					int y = p1[I[i].ind[1]] * m + p1[I[i].ind[3]];
					E2[x][y] += num_aln * I[i].x;
					i++;
				}
			}
			if (VERB)
				print_eta(N, ++count, size);
		}
	}
}

static void e2map_match2(struct index *I, int start, int end, int m, int n, double **E2) {
	unsigned int N = 2;
	if (m >= N && n >= N) {
		unsigned int a, b, size = round(expl(log_choose(m, N))), count = 0;

		if (VERB)
			print_eta(N, 0, size);
		for (a = 0; a < m; a++)
			for (b = a + 1; b < m; b++) {
				int i = start;
				unsigned int p1[] = { a, b, m };
				while (i <= end) {
					unsigned int p2[5] = { I[i].val[0], I[i].val[1], n };
					long double num_aln = numaln_fraction(p1, p2, N + 1);

					while (i <= end && SAME_DIST(I[i].val, p2, I[i].n)) {
						int x = p1[I[i].ind[0]] * m + p1[I[i].ind[2]];
						int y = p1[I[i].ind[1]] * m + p1[I[i].ind[3]];
						E2[x][y] += num_aln * I[i].x;

						if (p1[I[i].ind[2]] != p1[I[i].ind[3]]) {
							x = p1[I[i].ind[0]] * m + p1[I[i].ind[3]];
							y = p1[I[i].ind[1]] * m + p1[I[i].ind[2]];
							if (x < y)
								E2[x][y] += num_aln * I[i].x;
						}
						i++;
					}
				}
				if (VERB)
					print_eta(N, ++count, size);
			}
	}
}

static void e2map_match3(struct index *I, int start, int end, int m, int n, double **E2) {
	unsigned int N = 3;

	if (m >= N && n >= N) {
		unsigned int a, b, c, size = round(expl(log_choose(m, N))), count = 0;

		if (VERB)
			print_eta(N, 0, size);
		for (a = 0; a < m; a++)
			for (b = a + 1; b < m; b++)
				for (c = b + 1; c < m; c++) {
					int i = start;
					unsigned int p1[] = { a, b, c, m };
					while (i <= end) {
						unsigned int p2[5] = { I[i].val[0], I[i].val[1], I[i].val[2], n };
						long double num_aln = numaln_fraction(p1, p2, N + 1);

						while (i <= end && SAME_DIST(I[i].val, p2, I[i].n)) {
							int x = p1[I[i].ind[0]] * m + p1[I[i].ind[2]];
							int y = p1[I[i].ind[1]] * m + p1[I[i].ind[3]];
							E2[x][y] += num_aln * I[i].x;

							if (p1[I[i].ind[2]] != p1[I[i].ind[3]]) {
								x = p1[I[i].ind[0]] * m + p1[I[i].ind[3]];
								y = p1[I[i].ind[1]] * m + p1[I[i].ind[2]];
								if (x < y)
									E2[x][y] += num_aln * I[i].x;
							}
							i++;
						}
					}
					if (VERB)
						print_eta(N, ++count, size);
				}
	}
}

static void e2map_match4(struct index *I, int start, int end, int m, int n, double **E2) {
	unsigned int N = 4;

	if (m >= N && n >= N) {
		unsigned int a, b, c, d, size = round(expl(log_choose(m, N))), count = 0;

		if (VERB)
			print_eta(N, 0, size);
		for (a = 0; a < m; a++)
			for (b = a + 1; b < m; b++)
				for (c = b + 1; c < m; c++)
					for (d = c + 1; d < m; d++) {
						int i = start;
						unsigned int p1[] = { a, b, c, d, m };
						while (i <= end) {
							unsigned int p2[5] = { I[i].val[0], I[i].val[1], I[i].val[2], I[i].val[3], n };
							long double num_aln = numaln_fraction(p1, p2, N + 1);

							while (i <= end && SAME_DIST(I[i].val, p2, I[i].n)) {
								int x = p1[I[i].ind[0]] * m + p1[I[i].ind[2]];
								int y = p1[I[i].ind[1]] * m + p1[I[i].ind[3]];
								E2[x][y] += num_aln * I[i].x;

								if (p1[I[i].ind[2]] != p1[I[i].ind[3]]) {
									x = p1[I[i].ind[0]] * m + p1[I[i].ind[3]];
									y = p1[I[i].ind[1]] * m + p1[I[i].ind[2]];
									if (x < y)
										E2[x][y] += num_aln * I[i].x;
								}
								i++;
							}
						}
						if (VERB)
							print_eta(N, ++count, size);
					}
	}
}

static double cc_mean_perm_zero_diag(double **X, double **Y, unsigned int m, unsigned int n, unsigned int N) {
	double sumX2 = mat_sum2(X, m, 1);
	double sumY2 = mat_sum2(Y, n, 1);

	if (sumX2 * sumY2 == 0) {
		return 0;
	} else {
		double sumX = mat_sum(X, m, 1);
		double sumY = mat_sum(Y, n, 1);
		double L = N;

		return sumX * sumY / (L * (L - 1) * sqrt(sumX2 * sumY2));
	}
}

static double cc_mean_perm_nonzero_diag(double **X, double **Y, unsigned int m, unsigned int n, unsigned int N) {
	double sumX2 = mat_sum2(X, m, 0);
	double sumY2 = mat_sum2(Y, n, 0);

	if (sumX2 * sumY2 == 0) {
		return 0;
	} else {
		double trX = mat_trace(X, m);
		double trY = mat_trace(Y, n);
		double sumX = mat_sum(X, m, 1);
		double sumY = mat_sum(Y, n, 1);
		double L = N;

		double E1 = (trX * trY) / L;
		double E2 = sumX * sumY / (L * (L - 1));

		return (E1 + E2) / sqrt(sumX2 * sumY2);
	}
}

static double cc_var_perm_zero_diag(double **X, double **Y, unsigned int m, unsigned int n, unsigned int N) {
	double sumX2 = mat_sum2(X, m, 1);
	double sumY2 = mat_sum2(Y, n, 1);

	if (sumX2 * sumY2 == 0) {
		return 0;
	} else {
		double sumX = mat_sum(X, m, 1);
		double sumY = mat_sum(Y, n, 1);
		double sumXX = mat_sum_prod(X, X, m, 1, 1);
		double sumYY = mat_sum_prod(Y, Y, n, 1, 1);
		double L = N;

		double V1 = 2.0 / (L * (L - 1));
		double V2 = (4 * (sumXX - sumX2) * (sumYY - sumY2)) / (L * (L - 1) * (L - 2) * sumX2 * sumY2);
		double V3 = ((sumX * sumX - 4 * sumXX + 2 * sumX2) * (sumY * sumY - 4 * sumYY + 2 * sumY2)) / (L * (L - 1) * (L - 2) * (L - 3) * sumX2 * sumY2);
		double V4 = (sumX * sumX * sumY * sumY) / (L * L * (L - 1) * (L - 1) * sumX2 * sumY2);

		return V1 + V2 + V3 - V4;
	}
}

static double cc_var_perm_nonzero_diag(double **X, double **Y, unsigned int m, unsigned int n, unsigned int N) {
	double sumX2 = mat_sum2(X, m, 0);
	double sumY2 = mat_sum2(Y, n, 0);

	if (sumX2 * sumY2 == 0) {
		return 0;
	} else {
		double trX = mat_trace(X, m);
		double trY = mat_trace(Y, n);
		double trX2 = mat_trace2(X, m);
		double trY2 = mat_trace2(Y, n);
		double sumX = mat_sum(X, m, 0);
		double sumX0 = sumX - trX;
		double sumY = mat_sum(Y, n, 0);
		double sumY0 = sumY - trY;
		double sumX02 = mat_sum2(X, m, 1);
		double sumY02 = mat_sum2(Y, n, 1);
		double sumXX0 = mat_sum_prod(X, X, m, 0, 1);
		double sumX0X0 = mat_sum_prod(X, X, m, 1, 1);
		double sumYY0 = mat_sum_prod(Y, Y, n, 0, 1);
		double sumY0Y0 = mat_sum_prod(Y, Y, n, 1, 1);
		double L = N;

		double S11 = trX2 * trY2;
		double S21 = (trX * trX - trX2) * (trY * trY - trY2);
		double S22 = sumX02 * sumY02;
		double S23 = (sumXX0 - sumX0X0) * (sumYY0 - sumY0Y0);
		double S31 = (2 * sumX0X0 + sumX * sumX0 - 2 * sumXX0 - sumX0 * sumX0) * (2 * sumY0Y0 + sumY * sumY0 - 2 * sumYY0 - sumY0 * sumY0);
		double S32 = (sumX0X0 - sumX02) * (sumY0Y0 - sumY02);
		double S41 = (sumX0 * sumX0 - 4 * sumX0X0 + 2 * sumX02) * (sumY0 * sumY0 - 4 * sumY0Y0 + 2 * sumY02);

		double E, E2;

		E2 = S11 / L + (S21 + 2 * S22 + 4 * S23) / (L * (L - 1)) + (2 * S31 + 4 * S32) / (L * (L - 1) * (L - 2)) + S41 / (L * (L - 1) * (L - 2) * (L - 3));
		E = cc_mean_perm_nonzero_diag(X, Y, m, n, N);

		/*
		   printf("S1 = %f\n",S11/L);
		   printf("S2 = %f\n",(S21+2*S22+4*S23)/(L*(L-1)));
		   printf("S3 = %f\n",(2*S31+4*S32)/(L*(L-1)*(L-2)));
		   printf("S4 = %f\n",S41/(L*(L-1)*(L-2)*(L-3)));
		   printf("T  = %f\n",E2);
		   cc_mv_perm_check_bf(X,Y,m,n);
		 */

		return E2 / (sumX2 * sumY2) - E * E;
	}
}

static void cc_emap_bf_rec(double **Y, unsigned int m, unsigned int n, int *A, unsigned int mi, unsigned int ni, double *tot_aln, double **E) {
	static long double tot = 0;

	if (*tot_aln == 0 && tot == 0) {
		tot = expl(log_choose(m + n, MIN(m, n)));
		if (VERB)
			fprintf(stderr, "*** SETTING UP EMAP WITH BF ALGORITHM ***\n\n#Alignments: %LE\n\n", tot);
		if (VERB)
			print_eta(0, 0, tot);
	}

	if (mi < m) {
		int i;
		for (i = ni; i < n; i++) {
			A[mi] = i;
			cc_emap_bf_rec(Y, m, n, A, mi + 1, i + 1, tot_aln, E);
		}
		A[mi] = -1;
		cc_emap_bf_rec(Y, m, n, A, mi + 1, ni, tot_aln, E);
	} else if (mi == m) {
		double **T = mat_align(Y, A, m);

		mat_add(E, T, m);
		mat_double_free(T);
		*tot_aln += 1;

		if (VERB)
			print_eta(0, *tot_aln, tot);
		if (*tot_aln == tot) {
			tot = 0;
			if (VERB)
				fprintf(stderr, "\n");
		}
	}
}

void cc_e2map_bf_rec(double **Y, unsigned int m, unsigned int n, int *A, unsigned int mi, unsigned int ni, double *tot_aln, double **E2) {
	static long double tot = 0;

	if (*tot_aln == 0 && tot == 0) {
		tot = expl(log_choose(m + n, MIN(m, n)));
		if (VERB)
			fprintf(stderr, "*** SETTING UP E2MAP WITH BF ALGORITHM ***\n\n#Alignments: %LE\n\n", tot);
		if (VERB)
			print_eta(0, 0, tot);
	}

	if (mi < m) {
		int i;
		for (i = ni; i < n; i++) {
			A[mi] = i;
			cc_e2map_bf_rec(Y, m, n, A, mi + 1, i + 1, tot_aln, E2);
		}
		A[mi] = -1;
		cc_e2map_bf_rec(Y, m, n, A, mi + 1, ni, tot_aln, E2);
	} else if (mi == m) {
		double **T = mat_align(Y, A, m);
		double **K = mat_kronecker_prod(T, T, m);

		mat_add(E2, K, m * m);
		mat_double_free(T);
		mat_double_free(K);
		*tot_aln += 1;

		if (VERB)
			print_eta(0, *tot_aln, tot);
		if (*tot_aln == tot) {
			tot = 0;
			if (VERB)
				fprintf(stderr, "\n");
		}
	}
}

static int same_alignments(int *A, int *B, unsigned int m) {
	int i;
	for (i = 0; i < m; i++)
		if (A[i] != B[i])
			return 0;
	return 1;
}

static int *random_positions(unsigned int m, unsigned int k) {
	unsigned int i, tmp, newk, val;
	int *A = (int *)malloc(m * sizeof(int));

	if (k > m / 2) {
		for (i = 0; i < m; i++)
			A[i] = 1;
		newk = m - k;
		val = -1;
	} else {
		for (i = 0; i < m; i++)
			A[i] = -1;
		newk = k;
		val = 1;
	}

	i = 0;
	while (i != newk) {
		tmp = random() % m;
		if (A[tmp] != val) {
			A[tmp] = val;
			i++;
		}
	}

	return A;
}

static int *random_alignment(unsigned int m, unsigned int n, unsigned int k) {
	unsigned int iA, iB;
	int *A = random_positions(m, k);
	int *B = random_positions(n, k);

	iA = iB = 0;
	while (iA < m && iB < n) {
		while (iA < m && A[iA] == -1)
			iA++;
		while (iB < n && B[iB] == -1)
			iB++;
		if (iA < m && iB < n)
			A[iA++] = iB++;
	}

	free(B);
	return A;
}

static int **cc_mv_samp_alignments(unsigned int m, unsigned int n, unsigned int k, unsigned int num) {
	unsigned int i = 0, j, match;
	int **A = num == 0 ? NULL : (int **)calloc(num, sizeof(int *)), *tmp;

	while (i < num) {
		tmp = random_alignment(m, n, k);
		for (j = match = 0; j < i && match == 0; j++)
			match = same_alignments(A[j], tmp, m);
		if (match == 0)
			A[i++] = tmp;
		else
			free(tmp);
	}

	return A;
}

static void cc_mv_samp_free_alignments(int **A, unsigned int num) {
	unsigned int i;
	for (i = 0; i < num; i++)
		free(A[i]);
	free(A);
}

static void cc_mv_samp_add(double **X, double **Y, unsigned int m, int **A, unsigned int num, double *tmpE, double *tmpV) {
	unsigned int i;

	for (i = 0; i < num; i++) {
		double **T = mat_align(Y, A[i], m);
		double cc = mat_trace_prod(X, T, m);

		*tmpE += cc;
		*tmpV += cc * cc;
		free(T);
	}
}

static long double *cc_mv_samp_len_distr(unsigned int m, unsigned int n) {
	unsigned int len = MIN(m, n), k;
	long double *P = (long double *)calloc(len + 1, sizeof(long double));
	long double log_tot_aln = log_choose(m + n, MIN(m, n));

	for (k = 0; k <= len; k++)
		P[k] = expl(log_choose(m, k) + log_choose(n, k) - log_tot_aln);

	return P;
}

static long double log_fact(unsigned int n) {
	long double x = 0;
	while (n > 1)
		x += logl(n--);
	return x;
}

static long double log_choose(unsigned int m, unsigned int k) {
	unsigned int i;
	long double x = 0;

	for (i = 0; i < k; i++)
		x += logl(m - i);

	return x - log_fact(k);
}

static void logbin_matrix_setup(unsigned int m, unsigned int n) {
	unsigned int i, j;

	LOGBIN = mat_ldouble_alloc(m + 1, n + 1);
	for (i = 0; i <= m; i++)
		for (j = 0; j <= n; j++)
			LOGBIN[i][j] = log_choose(i + j, MIN(i, j));
}

static void logbin_matrix_free(void) {
	free(LOGBIN[0]);
	free(LOGBIN);
	LOGBIN = NULL;
}

static long double numaln_fraction(unsigned int p1[], unsigned int p2[], unsigned int n) {
	long double tmp = 0;

	if (n > 1) {
		int i, x1, x2;

		for (i = 0; i < n; i++) {
			if (i == 0) {
				x1 = p1[i];
				x2 = p2[i];
			} else {
				x1 = p1[i] - p1[i - 1] - 1;
				x2 = p2[i] - p2[i - 1] - 1;
			}
			tmp += LOGBIN[x1][x2];
		}
	}

	return expl(tmp - LOGBIN[p1[n - 1]][p2[n - 1]]);
}

static long double numaln_fraction_match1(unsigned int i, unsigned int k, unsigned int m, unsigned int n) {
	unsigned int s1 = i, s2 = k;
	unsigned int e1 = m - i - 1, e2 = n - k - 1;

	return expl(LOGBIN[s1][s2] + LOGBIN[e1][e2] - LOGBIN[m][n]);
}

static long double numaln_fraction_match2(unsigned int i, unsigned int j, unsigned int k, unsigned int l, unsigned int m, unsigned int n) {
	unsigned int s1 = i, s2 = k;
	unsigned int c1 = j - i - 1, c2 = l - k - 1;
	unsigned int e1 = m - j - 1, e2 = n - l - 1;

	return expl(LOGBIN[s1][s2] + LOGBIN[c1][c2] + LOGBIN[e1][e2] - LOGBIN[m][n]);
}

static void heap_perm(double **X, double **Y, unsigned int m, unsigned int n, int *P, unsigned int k, double *tot_perm, double *tot_cc, double *tot_cc2) {
	static long double tot = 0;

	if (*tot_perm == 0 && tot == 0) {
		tot = expl(log_fact(m + n));
		if (VERB)
			fprintf(stderr, "*** SETTING UP PERMUTATION MEAN AND VAR WITH BF ALGORITHM ***\n\n#Permutations: %LE\n\n", tot);
		if (VERB)
			print_eta(0, 0, tot);
	}

	if (k == 1) {
		double cc = cc_aln(X, Y, m, n, P);

		*tot_cc += cc;
		*tot_cc2 += cc * cc;
		*tot_perm += 1;

		if (VERB)
			print_eta(0, *tot_perm, tot);
		if (*tot_perm == tot) {
			tot = 0;
			if (VERB)
				fprintf(stderr, "\n");
		}

	} else {
		unsigned int i;
		// Generate permutations with kth unaltered
		// Initially k == length(A) = m+n
		heap_perm(X, Y, m, n, P, k - 1, tot_perm, tot_cc, tot_cc2);

		// Generate permutations for kth swapped with each k-1 initial
		for (i = 0; i < k - 1; i++) {
			// Swap choice dependent on parity of k (even or odd)
			if (k % 2 == 0) {
				SWAP(P, i, k - 1);
			} else {
				SWAP(P, 0, k - 1);
			}
			heap_perm(X, Y, m, n, P, k - 1, tot_perm, tot_cc, tot_cc2);

		}
	}
}

int distinct(int arr[], int n) {
	int i, res = 1;

	// Pick all elements one by one 
	for (i = 1; i < n; i++) {
		int j = 0;
		for (j = 0; j < i; j++)
			if (arr[i] == arr[j])
				break;

		// If not printed earlier, then print it 
		if (i == j)
			res++;
	}
	return res;
}

static void heap_perm2(double **X, double **Y, unsigned int m, unsigned int n, int *P, unsigned int k, double *tot_perm, double *S1, double *S2, double *S3, double *S4) {
	static long double tot = 0;

	if (*tot_perm == 0 && tot == 0) {
		tot = expl(log_fact(m + n));
		if (VERB)
			fprintf(stderr, "*** SETTING UP PERMUTATION MEAN AND VAR WITH BF ALGORITHM ***\n\n#Permutations: %LE\n\n", tot);
		if (VERB)
			print_eta(0, 0, tot);
	}

	if (k == 1) {
		double **YA = mat_align(Y, P, m);
		unsigned int i, j, k, l;

		for (i = 0; i < m; i++)
			for (j = 0; j < m; j++)
				for (k = 0; k < m; k++)
					for (l = 0; l < m; l++) {
						int a[] = { i, j, k, l };
						switch (distinct(a, 4)) {
						case 1:
							*S1 += X[i][j] * X[k][l] * YA[i][j] * YA[k][l];
							break;
						case 2:
							*S2 += X[i][j] * X[k][l] * YA[i][j] * YA[k][l];
							break;
						case 3:
							*S3 += X[i][j] * X[k][l] * YA[i][j] * YA[k][l];
							break;
						case 4:
							*S4 += X[i][j] * X[k][l] * YA[i][j] * YA[k][l];
							break;
						}
					}

		*tot_perm += 1;

		mat_double_free(YA);

		if (VERB)
			print_eta(0, *tot_perm, tot);
		if (*tot_perm == tot) {
			tot = 0;
			if (VERB)
				fprintf(stderr, "\n");
		}

	} else {
		unsigned int i;
		// Generate permutations with kth unaltered
		// Initially k == length(A) = m+n
		heap_perm2(X, Y, m, n, P, k - 1, tot_perm, S1, S2, S3, S4);

		// Generate permutations for kth swapped with each k-1 initial
		for (i = 0; i < k - 1; i++) {
			// Swap choice dependent on parity of k (even or odd)
			if (k % 2 == 0) {
				SWAP(P, i, k - 1);
			} else {
				SWAP(P, 0, k - 1);
			}
			heap_perm2(X, Y, m, n, P, k - 1, tot_perm, S1, S2, S3, S4);

		}
	}
}

void cc_mv_perm_check_bf(double **X, double **Y, unsigned int m, unsigned int n) {
	int i, *P = (int *)calloc(m + n, sizeof(int));
	double **YE = mat_extend(Y, n, m + n);
	double tot_perm = 0, S1 = 0, S2 = 0, S3 = 0, S4 = 0;

	for (i = 0; i < m + n; i++)
		P[i] = i;

	heap_perm2(X, YE, m, n, P, m + n, &tot_perm, &S1, &S2, &S3, &S4);

	printf("S1 = %Lf*\n", S1 / expl(log_fact(m + n)));
	printf("S2 = %Lf*\n", S2 / expl(log_fact(m + n)));
	printf("S3 = %Lf*\n", S3 / expl(log_fact(m + n)));
	printf("S4 = %Lf*\n", S4 / expl(log_fact(m + n)));
	printf("T = %LF*\n", (S1 + S2 + S3 + S4) / expl(log_fact(m + n)));

	mat_double_free(YE);
	free(P);
}

static void cc_mv_bf_rec(double **X, double **Y, unsigned int m, unsigned int n, int *A, unsigned int mi, unsigned int ni, double *tot_aln, double *tot_cc, double *tot_cc2) {
	static long double tot = 0;

	if (*tot_aln == 0 && tot == 0) {
		tot = expl(log_choose(m + n, MIN(m, n)));
		if (VERB)
			fprintf(stderr, "*** SETTING UP MEAN AND VAR WITH BF ALGORITHM ***\n\n#Alignments: %LE\n\n", tot);
		if (VERB)
			print_eta(0, 0, tot);
	}

	if (mi < m) {
		int i;
		for (i = ni; i < n; i++) {
			A[mi] = i;
			cc_mv_bf_rec(X, Y, m, n, A, mi + 1, i + 1, tot_aln, tot_cc, tot_cc2);
		}
		A[mi] = -1;
		cc_mv_bf_rec(X, Y, m, n, A, mi + 1, ni, tot_aln, tot_cc, tot_cc2);
	} else if (mi == m) {
		double cc = cc_aln(X, Y, m, n, A);

		*tot_cc += cc;
		*tot_cc2 += cc * cc;
		*tot_aln += 1;

		if (VERB)
			print_eta(0, *tot_aln, tot);
		if (*tot_aln == tot) {
			tot = 0;
			if (VERB)
				fprintf(stderr, "\n");
		}
	}
}

inline static char *ETA(clock_t start_t, clock_t end_t, unsigned int count, unsigned int size) {
	static char buf[100] = { 0 };
	double curr_time = 1.0 * (end_t - start_t) / CLOCKS_PER_SEC;
	double time = curr_time * (size - count) / count;
	unsigned int delay = 100;

	if (count % delay == 1) {
		if (time < 60)
			sprintf(buf, "%2d.%02ds", (int)time, (int)(100 * (time - (int)time)));
		else if (time < 3600)
			sprintf(buf, "%2d.%02dm", (int)time / 60, (int)time % 60);
		else if (time < 3600 * 24)
			sprintf(buf, "%2d.%02dh", (int)time / 3600, ((int)time % 3600) / 60);
		else if (time < 3600 * 24 * 7)
			sprintf(buf, "%2d.%02dd", (int)time / (3600 * 24), (int)time % (3600 * 24) / 3600);
		else if (time < 3600 * 24 * 30)
			sprintf(buf, "%.2fw", time / (3600 * 24 * 7));
		else if (time < 3600 * 24 * 365)
			sprintf(buf, "%.2fM", time / (3600 * 24 * 30));
		else
			sprintf(buf, "%.2fY", time / (3600 * 24 * 365));
	}
	return buf;
}

static void print_eta(unsigned int N, unsigned int count, unsigned int size) {
	static time_t start_t = 0;
	if (count == 0) {
		start_t = clock();
	} else {
		time_t end_t = clock();
		char buf[6];
		fprintf(stderr, "\x1b[?25l");
		sprintf(buf, "%3.1f", 100.0 * (count) / size);
		if (N == 0)
			//fprintf(stderr, "\r- Computing all matches. %5s%% ETA: %-7s %u/%u", buf, ETA(start_t, end_t, count, size), count, size);
			fprintf(stderr, "\r- Computing all matches. %5s%% ETA: %-7s", buf, ETA(start_t, end_t, count, size));
		else
			//fprintf(stderr, "\r- Computing %d matches. %5s%% ETA: %-7s %u/%u", N, buf, ETA(start_t, end_t, count, size), count, size);
			fprintf(stderr, "\r- Computing %d matches. %5s%% ETA: %-7s", N, buf, ETA(start_t, end_t, count, size));
		fprintf(stderr, "\x1b[?25h");
		fflush(NULL);
		if (count == size)
			fprintf(stderr, "\n");
	}
}
