/*********************************************************************
This file is part of the ccpro-1.0 software. 

Copyright (C) 2020  Pietro Di Lena

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
**********************************************************************/

#ifndef CCSTAT_H
#define CCSTAT_H

// Computes the expectation matrix E[Y^alpha]/sqrt(tr(YY)) of size m x m,
// over all alignments alpha : {1,..,m} -> {1,..,n}.
// Y is a n x n simmetric matrix. 
double **cc_emap(double **Y, unsigned int m, unsigned int n);

// Computes the expectation matrix E[Y^alpha * Y^alpha]/tr(YY) of size m^2 x m ^2 
// over all alignments alpha : {1,..,m} -> {1,..,n}, where * is the
// Kronecked product. Y is a n x n symmetric matrix. 
double **cc_e2map(double **Y, unsigned int m, unsigned int n);

// Computes the variance-covariance matrix Var[Y^alpha]/tr(YY) of size m^2 x m ^2 
// over all alignments alpha : {1,..,m} -> {1,..,n}, where
// Var[Y^alpha] = E[Y^alpha * Y^alpha] - E[Y^alpha] * E[Y^alpha] and
// * is the Kronecked product. Y is a n x n simmetric matrix
double **cc_vmap(double **Y, unsigned int m, unsigned int n);

// Computes the expected congruence coefficient E[cc^alpha(X,Y)] = XE[Y^alpha]/sqrt(tr(XX)tr(YY))
// over all alignments alpha : {1,..,m} -> {1,..,n}.
// - X is a m x m symmetric matrix
// - Y is a n x n symmetric matrix
// - E is the expectation matrix E[Y^alpha]/sqrt(tr(YY)). If E == NULL, it is computed
//   within the procedure from Y and m.
double cc_mean(double **X, double **Y, unsigned int m, unsigned int n, double **E);

// Computes the expected congruence coefficient E[cc^sigma(X,Y)] over all permutations
// sigma : {1,..,m+n} -> {1,..,m+n}
double cc_mean_perm(double **X, double **Y, unsigned int m, unsigned int n);

// Computes the variance o the congruence coefficient E[cc^alpha(X,Y)] = (X*X)Var[Y^alpha]/(tr(XX)tr(YY))
// over all alignments alpha : {1,..,m} -> {1,..,n}.
// - X is a m x m symmetric matrix
// - Y is a n x n symmetric matrix
// - V is the variance covariance matrix Var[Y^alpha]/tr(YY). If V = NULL, the procedure
//   considers matrices E and E2 for variance computation
// - E is the expectation matrix E[Y^alpha]/sqrt(tr(YY)). If V == NULL and E == NULL, E is computed
//   within the procedure from Y and m.
// - E2 is the the expectation matrix E[Y^alpha * Y^alpha]/tr(YY). If V == NULL and E2 == NULL, E2
//   is computed withing the procedure from Y and m.
double cc_var(double **X, double **Y, unsigned int m, unsigned int n, double **E, double **E2, double **V);

// Computes the variance of the congruence coefficient E[cc^sigma(X,Y)] over all permutations
// sigma : {1,..,m+n} -> {1,..,m+n}
double cc_var_perm(double **X, double **Y, unsigned int m, unsigned int n);

// Computes mean and variance of the congruence coefficient by randomly sampling alignments.
// - X is a m x m symmetric matrix
// - Y is a n x n symmetric matrix
// - samp_size is the number of alignments to be sampled
// Returns the actual number of sampled alignments (0 if either X or Y are zero matrices). 
unsigned int cc_mv_samp(double **X, double **Y, unsigned int m, unsigned int n, unsigned int samp_size, double *mean, double *var);

// congruence coefficient between two symmetric matrices of size m x m
double cc(double **X, double **Y, unsigned int m);

// congruence coefficient between two general matrices of size m x n
double cc_gen(double **X, double **Y, unsigned int m, unsigned int n);

// congruence coefficient between two symmetric matrices of size m x m (X)
// and n x n (Y) with respect to alignment A
double cc_aln(double **X, double **Y, unsigned int m, unsigned int n, int *A);

// Right-tailed p-value for the cc coefficient between two (aligned) symmetric 
// matrices X of size m x m and Y of size n x n.
// The defree of freedom df is the number of non zero elements in the upper
// triangular portion of one of the two matrices.
double cc_pval(double cc, unsigned int df);

// Right-tailed p-value for the cc coefficient between two general matrices
// X,Y of size m x n, where the degree of freedom df = mn.
double cc_pval_gen(double cc, unsigned int df);

// Set verbose on for emaps calculations
void cc_verbose_on(void);

// Set verbose off for emaps calculations
void cc_verbose_off(void);

/*****************************
        TESTING FUNCTIONS
 *****************************/

// Functions used for debugging. Can be used only for small
// sizes m and n

// Brute force computation of the expectation matrix.
double **cc_emap_bf(double **Y, unsigned m, unsigned n);

// Brute force computation of the expectation2 matrix.
double **cc_e2map_bf(double **Y, unsigned m, unsigned n);

// Brute force computation of the variance-covariance matrix.
double **cc_vmap_bf(double **Y, unsigned m, unsigned n);

// Brute force computation of mean and variance over all possible alignments
void cc_mv_bf(double **X, double **Y, unsigned int m, unsigned int n, double *mean, double *var);

// Brute force computation of mean and variance over all possible permutations
void cc_mv_perm_bf(double **X, double **Y, unsigned int m, unsigned int n, double *mean, double *var);

void cc_mv_perm_check_bf(double **X, double **Y, unsigned int m, unsigned int n);

#endif
