/*********************************************************************
This file is part of the ccpro-1.0 software. 

Copyright (C) 2020  Pietro Di Lena

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <ctype.h>
#include "mat.h"

#define MAX_FASTA 10000

// GLOBAL OPTIONS
static char *ALN = NULL;	// Fasta file for casp format
static char *TYPE = NULL;	// Output type: casp or simp
static char *FASTA1 = NULL;	// First fasta sequence
static char *FASTA2 = NULL;	// Second fasta sequence

static char *getname(char *buf) {
	unsigned int i, j, n = strlen(buf);
	char *name = (char *)calloc(n, sizeof(char));

	//printf("%s\n",buf);
	i = 0;
	// Skip header symbol and balnk symbols
	while (buf[i] != '\0' && (buf[i] == '>' || isblank(buf[i])))
		i++;
	// Copy header name/description
	for (j = 0; buf[i] != '\0'; i++)
		name[j++] = buf[i];
	return name;
}

static char *read_fasta(char *file, char **name) {
	FILE *in;
	char *line = NULL, *fasta = NULL;
	size_t len = 0;
	ssize_t read;

	if ((in = fopen(file, "r")) != NULL) {
		unsigned int curr_size = MAX_FASTA, check = 0;
		fasta = (char *)malloc((curr_size + 1) * sizeof(char));
		while ((read = getline(&line, &len, in)) != -1 && (check == 0 || line[0] != '>')) {
			// Delete trailing newline
			if (line[read - 1] == '\n')
				line[read - 1] = '\0';
			// Skip header
			if (line[0] != '>') {
				if (strlen(fasta) + strlen(line) > curr_size) {
					// Warning: doesn't deal with reallocation errors
					curr_size = 2 * (strlen(fasta) + strlen(line)) + 1;
					fasta = (char *)realloc(fasta, curr_size);
				}
				strncat(fasta, line, strlen(line));
			} else {
				*name = getname(line);
				check = 1;
			}
		}
		free(line);
		fasta = (char *)realloc(fasta, strlen(fasta) + 1);
	}

	return fasta;
}

static void usage(void) {
	fprintf(stderr, "Usage: convert_aln [options] <file> <fasta|simp>\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "\t-1, --fasta1=<file>  First fasta sequence for fasta format output.\n");
	fprintf(stderr, "\t-2, --fasta2=<file>  Second fasta sequence for fasta format output.\n");
	fprintf(stderr, "\t-h, --help           Print this help\n");
}

static int parse_args(int argc, char *argv[]) {
	int opt = 0, long_index = 0;

	static struct option long_options[] = {
		{"fasta1", required_argument, 0, '1'},
		{"fasta2", required_argument, 0, '2'},
		{"help", no_argument, 0, 'h'},
		{0, 0, 0, 0}
	};

	while ((opt = getopt_long(argc, argv, "1:2:h", long_options, &long_index)) != -1) {
		switch (opt) {
		case '1':
			FASTA1 = optarg;
			break;
		case '2':
			FASTA2 = optarg;
			break;
		case 'h':
			usage();
			return 0;
		default:
			fprintf(stderr, "Unrecognized option -%c\n\n", opt);
			return 1;
		}
	}

	if (argc - optind < 2) {
		usage();
		return 1;
	} else if (argc - optind > 2) {
		int i;
		fprintf(stderr, "Unrecognized arguments:");
		for (i = optind + 2; i < argc; i++)
			fprintf(stderr, " %s", argv[i]);
		fprintf(stderr, "\n\n");
		return 1;
	}

	ALN = argv[optind];
	TYPE = argv[optind + 1];

	return 0;
}

static void print_fasta_aln(int *A, char *fasta1, char *fasta2, char *name1, char *name2) {
	unsigned int n = strlen(fasta1);
	unsigned int m = strlen(fasta2);
	int i, j, curr = 0;

	printf(">%s\n", name1 == NULL ? "FIRST SEQUENCE" : name1);
	i = curr = 0;
	while (i < n) {
		if (A[i] == -1) {
			printf("%c", fasta1[i]);
		} else {
			for (j = curr; j < A[i]; j++)
				printf("-");
			printf("%c", fasta1[i]);
			curr = A[i] + 1;
		}
		i++;
	}
	for (j = curr; j < m; j++)
		printf("-");
	printf("\n");
	printf(">%s\n", name2 == NULL ? "SECOND SEQUENCE" : name2);
	i = curr = 0;
	while (i < n) {
		if (A[i] == -1) {
			printf("-");
		} else {
			for (j = curr; j <= A[i]; j++)
				printf("%c", fasta2[j]);
			curr = A[i] + 1;
		}
		i++;
	}
	for (j = curr; j < m; j++)
		printf("%c", fasta2[j]);
	printf("\n");
}

static void print_simp_aln(int *A, unsigned int m, unsigned int n) {
	unsigned int i;
	printf("%u %u\n", m, n);
	for (i = 0; i < m; i++)
		if (A[i] != -1)
			printf("%d:%d\n", i, A[i]);
}

int main(int argc, char *argv[]) {
	int *A;
	char *fasta1 = NULL, *fasta2 = NULL, *name1, *name2;
	unsigned int m, n;

	if (parse_args(argc, argv) != 0)
		return 1;

	if ((A = mat_aln_read(ALN, &m, &n)) == NULL) {
		fprintf(stderr, "Error reading %s\n", ALN);
		return 1;
	}

	if (strcmp(TYPE, "fasta") == 0) {
		if ((fasta1 = read_fasta(FASTA1, &name1)) == NULL) {
			fprintf(stderr, "Error reading fasta sequence in %s file\n", FASTA1);
			return 1;
		}
		if (strlen(fasta1) != m) {
			fprintf(stderr, "Error: length mismatch between alignment and first fasta sequence\n");
			return 1;
		}
		if ((fasta2 = read_fasta(FASTA2, &name2)) == NULL) {
			fprintf(stderr, "Error reading fasta sequence in %s file\n", FASTA2);
			return 1;
		}
		if (strlen(fasta2) != n) {
			fprintf(stderr, "Error: length mismatch between alignment and second fasta sequence\n");
			return 1;
		}
		print_fasta_aln(A, fasta1, fasta2, name1, name2);
	} else if (strcmp(TYPE, "simp") == 0) {
		print_simp_aln(A, m, n);
	} else {
		fprintf(stderr, "Error: output format should be either fasta or simp\n");
		return 1;
	}

	return 0;
}
