/*********************************************************************
This file is part of the ccpro-1.0 software. 

Copyright (C) 2020  Pietro Di Lena

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include "mat.h"

#define MAX_FASTA 10000

// GLOBAL OPTIONS
static char *FASTA = NULL;	// Fasta file for casp format
static char *NAME = NULL;	// Target name for casp format
static char *TYPE = NULL;	// Output type: casp or simp
static char *MAP = NULL;	// Input map
static unsigned int MOD = 1;	// Model number for casp format

static int read_ui(char *num, unsigned int *n) {
	char *ptr;
	int base = 10;

	*n = strtoul(num, &ptr, base);
	return num == ptr || errno != 0 || (errno == 0 && num && *ptr != 0);
}

static char *read_fasta(char *file) {
	FILE *in;
	char *line = NULL, *fasta = NULL;
	size_t len = 0;
	ssize_t read;

	if ((in = fopen(file, "r")) != NULL) {
		unsigned int curr_size = MAX_FASTA, check = 0;
		fasta = (char *)malloc((curr_size + 1) * sizeof(char));
		while ((read = getline(&line, &len, in)) != -1 && (check == 0 || line[0] != '>')) {
			// Delete trailing newline
			if (line[read - 1] == '\n')
				line[read - 1] = '\0';
			// Skip header
			if (line[0] != '>') {
				if (strlen(fasta) + strlen(line) > curr_size) {
					// Warning: doesn't deal with reallocation errors
					curr_size = 2 * (strlen(fasta) + strlen(line)) + 1;
					fasta = (char *)realloc(fasta, curr_size);
				}
				strncat(fasta, line, strlen(line));
			} else {
				check = 1;
			}
		}
		free(line);
		fasta = (char *)realloc(fasta, strlen(fasta) + 1);
	}

	return fasta;
}

static int cmp(const void *a, const void *b) {
	double *x = *(double **)a;
	double *y = *(double **)b;

	//fprintf(stderr,"x=%f y=%f\n",x[2],y[2]);
	if (x[2] == y[2])
		return x[1] == y[1] ? x[2] - y[2] : x[1] - x[1];
	else
		return x[2] < y[2] ? 1 : -1;
}

static double **mat_to_sorted_list(double **M, unsigned int n, unsigned int *N) {
	double **list = mat_to_list(M, n, N);

	// We should save list[0] ptr in order to free list later
	qsort(list, *N, sizeof(double *), cmp);

	return list;
}

static void print_casp(double **M, char *fasta, unsigned int n, char *name, unsigned int model) {
	unsigned int N = 0, i;
	double **L = mat_to_sorted_list(M, n, &N);
	printf("PFRMAT RR\n");
	printf("TARGET %s\n", name == NULL ? "NONAME" : name);
	printf("MODEL %u\n", model);
	for (i = 0; i < n; i++) {
		printf("%c", fasta[i]);
		if ((i + 1) % 50 == 0 && i + 1 < n)
			printf("\n");
	}
	printf("\n");
	for (i = 0; i < N; i++)
		if (L[2] != 0)
			printf("%u %u 0 8 %f\n", (unsigned int)L[i][0] + 1, (unsigned int)L[i][1] + 1, L[i][2]);
	printf("END\n");
}

static void print_simp(double **M, unsigned int n) {
	unsigned int N = 0, i;
	double **L = mat_to_sorted_list(M, n, &N);

	printf("%u\n", n);
	for (i = 0; i < N; i++)
		if (L[2] != 0)
			printf("%u %u %f\n", (unsigned int)L[i][0], (unsigned int)L[i][1], L[i][2]);
}

void usage(void) {
	fprintf(stderr, "Usage: convert_map [options] <file> <casp|simp>\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "\t-f, --fasta=<file>   Fasta sequence for casp format output\n");
	fprintf(stderr, "\t-n, --name=<name>    Target name for casp format output\n");
	fprintf(stderr, "\t-m, --model=<int>    Model number for casp format output\n");
	fprintf(stderr, "\t-h, --help           Print this help\n");
}

int parse_args(int argc, char *argv[]) {
	int opt = 0, long_index = 0;

	static struct option long_options[] = {
		{"fasta", required_argument, 0, 'f'},
		{"name", required_argument, 0, 'n'},
		{"model", required_argument, 0, 'm'},
		{"help", no_argument, 0, 'h'},
		{0, 0, 0, 0}
	};

	while ((opt = getopt_long(argc, argv, "f:n:m:h", long_options, &long_index)) != -1) {
		switch (opt) {
		case 'f':
			FASTA = optarg;
			break;
		case 'n':
			NAME = optarg;
			break;
		case 'm':
			if (read_ui(optarg, &MOD) != 0) {
				fprintf(stderr, "Error converting -m argument: %s\n", optarg);
				return 1;
			}
			break;
		case 'h':
			usage();
			return 0;
		default:
			fprintf(stderr, "Unrecognized option -%c\n\n", opt);
			return 1;
		}
	}

	if (argc - optind < 2) {
		usage();
		return 1;
	} else if (argc - optind > 2) {
		int i;
		fprintf(stderr, "Unrecognized arguments:");
		for (i = optind + 2; i < argc; i++)
			fprintf(stderr, " %s", argv[i]);
		fprintf(stderr, "\n\n");
		return 1;
	}

	MAP = argv[optind];
	TYPE = argv[optind + 1];

	return 0;
}

int main(int argc, char *argv[]) {
	double **M;
	char *fasta = NULL;
	unsigned int n;

	if (parse_args(argc, argv) != 0)
		return 1;

	if ((M = mat_read(MAP, &n, 0)) == NULL) {
		fprintf(stderr, "Error reading %s\n", MAP);
		return 1;
	}

	if (strcmp(TYPE, "casp") == 0) {
		if ((fasta = read_fasta(FASTA)) == NULL) {
			fprintf(stderr, "Error: fasta sequence not provided or unreadable\n");
			return 1;
		}
		if (strlen(fasta) != n) {
			fprintf(stderr, "Error: inconsistency between sequence and map lengths\n");
			return 1;
		}
		print_casp(M, fasta, n, NAME, MOD);
	} else if (strcmp(TYPE, "simp") == 0) {
		print_simp(M, n);
	} else {
		fprintf(stderr, "Error: output format should be either casp or simp\n");
		return 1;
	}

	return 0;
}
