/*********************************************************************
This file is part of the ccpro-1.0 software. 

Copyright (C) 2020  Pietro Di Lena

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "mat.h"

#define SIMP_FORMAT  0
#define CASP_FORMAT  1
#define FASTA_FORMAT 2
#define MAX_FASTA   10000

#define ABS(x) ((x) > 0 ? x : -(x))
#define ISGZ(file) (strcmp(file+strlen(file)-4,".gz") == 0)
#define CMD(cmd,file) strncat(strncpy(cmd,"zcat ",5),file,strlen(file))

/**************************************
        PRIVATE FUNCTIONS PROTOTYPES
 **************************************/

// Checks type: SIMP_FORMAT | CASP_FORMAT | FASTA_FORMAT
static int mat_read_type(char *file);

// Read matrix simple format
static double **mat_read_simple(char *file, unsigned int *n, unsigned int diag);

// Read casp simple format
static double **mat_read_casp(char *file, unsigned int *n, unsigned int diag);

// Read alignment simple format
static int *mat_aln_read_simp(char *file, unsigned int *m, unsigned int *n);

// Read alignment fasta format
static int *mat_aln_read_fasta(char *file, unsigned int *m, unsigned int *n, char **fasta1, char **fasta2);

/******************************
        MATRIX ALLOCATION
 ******************************/

double **mat_double_alloc(unsigned int row, unsigned int col) {
	double **M = (double **)malloc(row * sizeof(double *));
	unsigned int i;

	if (M != NULL) {
		M[0] = (double *)calloc(row * col, sizeof(double));
		if (M[0] == NULL) {
			free(M);
			M = NULL;
		} else {
			for (i = 0; i < row; i++)
				M[i] = &((M[0])[col * i]);
		}
	}
	return M;
}

long double **mat_ldouble_alloc(unsigned int row, unsigned int col) {
	long double **M = (long double **)malloc(row * sizeof(long double *));
	unsigned int i;

	if (M != NULL) {
		M[0] = (long double *)calloc(row * col, sizeof(long double));
		if (M[0] == NULL) {
			free(M);
			M = NULL;
		} else {
			for (i = 0; i < row; i++)
				M[i] = &((M[0])[col * i]);
		}
	}
	return M;
}

void mat_double_free(double **M) {
	if (M != NULL)
		free(M[0]);
	free(M);
}

/******************************
       matrix IO 
 ******************************/

double **mat_read(char *file, unsigned int *n, unsigned int diag) {
	int type = mat_read_type(file);
	double **M = NULL;

	if (type == CASP_FORMAT)
		M = mat_read_casp(file, n, diag);
	else if (type == SIMP_FORMAT)
		M = mat_read_simple(file, n, diag);

	return M;
}

int *mat_aln_read(char *file, unsigned int *m, unsigned int *n) {
	int type = mat_read_type(file);
	int *A = NULL;
	char *fasta1 = NULL, *fasta2 = NULL;

	if (type == FASTA_FORMAT)
		A = mat_aln_read_fasta(file, m, n, &fasta1, &fasta2);
	else if (type == SIMP_FORMAT)
		A = mat_aln_read_simp(file, m, n);

	return A;
}

static void _mat_save(double **M, unsigned int n, FILE * out) {
	unsigned int i, j;
	fprintf(out, "%u\n", n);
	for (i = 0; i < n; i++) {
		for (j = i; j < n; j++)
			if (M[i][j] != 0)
				fprintf(out, "%u %u %g\n", i, j, M[i][j]);
	}
}

int mat_save(double **M, unsigned int n, char *filename) {
	FILE *out;
	if ((out = fopen(filename, "w")) == NULL) {
		return 0;
	} else {
		_mat_save(M, n, out);
		return 1;
	}
}

void mat_print(double **M, unsigned int n) {
	_mat_save(M, n, stdout);
}

void mat_full_print(double **M, unsigned int n) {
	unsigned int i, j;
	printf("%u\n", n);
	for (i = 0; i < n; i++) {
		printf("%.4g", M[i][0]);
		for (j = 1; j < n; j++)
			printf(" %.4g", M[i][j]);
		printf("\n");
	}
}

/******************************
       Matrix operation
 ******************************/

int mat_zero_diag(double **X, unsigned int n) {
	if (X == NULL) {
		return 0;
	} else {
		unsigned int i;
		for (i = 0; i < n; i++)
			if (X[i][i] != 0)
				return 0;
		return 1;
	}
}

double mat_trace(double **X, unsigned int n) {
	if (X == NULL) {
		return 0;
	} else {
		unsigned int i;
		double tr = 0;
		for (i = 0; i < n; i++)
			tr += X[i][i];
		return tr;
	}
}

double mat_trace2(double **X, unsigned int n) {
	if (X == NULL) {
		return 0;
	} else {
		unsigned int i;
		double tr = 0;
		for (i = 0; i < n; i++)
			tr += X[i][i] * X[i][i];

		return tr;
	}
}

double mat_trace_prod(double **X, double **Y, unsigned int n) {
	if (X == NULL || Y == NULL) {
		return 0;
	} else {
		unsigned int i, j;
		double tr = 0;
		for (i = 0; i < n; i++)
			for (j = 0; j < n; j++)
				tr += X[i][j] * Y[i][j];
		return tr;
	}
}

double mat_trace_prod_gen(double **X, double **Y, unsigned int m, unsigned int n) {
	if (X == NULL || Y == NULL) {
		return 0;
	} else {
		unsigned int i, j;
		double tr = 0;
		for (i = 0; i < n; i++)
			for (j = 0; j < n; j++)
				tr += X[i][j] * Y[j][i];
		return tr;
	}
}

double mat_trace_prod_aln(double **X, double **Y, int *A, unsigned int n) {
	if (X == NULL || Y == NULL || A == NULL) {
		return 0;
	} else {
		unsigned int i, j;
		double tr = 0;
		for (i = 0; i < n; i++)
			if (A[i] != -1)
				for (j = 0; j < n; j++)
					if (A[j] != -1)
						tr += X[i][j] * Y[A[i]][A[j]];
		return tr;
	}
}

void mat_round(double **X, unsigned int n, double cutoff) {
	if (X != NULL) {
		unsigned int i, j;
		for (i = 0; i < n; i++)
			for (j = 0; j < n; j++)
				if (ABS(X[i][j]) < cutoff)
					X[i][j] = 0;
	}
}

double **mat_kronecker_prod(double **X, double **Y, unsigned int n) {
	if (X == NULL || Y == NULL) {
		return NULL;
	} else {
		double **K = mat_double_alloc(n * n, n * n);
		unsigned i, j, k, l;

		if (K != NULL) {
			for (i = 0; i < n; i++)
				for (j = 0; j < n; j++)
					for (k = 0; k < n; k++)
						for (l = 0; l < n; l++)
							K[i * n + j][k * n + l] = X[i][k] * Y[j][l];
		}

		return K;
	}
}

double **mat_align(double **X, int *A, unsigned int n) {
	if (X == NULL || A == NULL) {
		return NULL;
	} else {
		double **Y = mat_double_alloc(n, n);
		unsigned int i, j;

		if (Y != NULL) {
			for (i = 0; i < n; i++)
				for (j = i; j < n; j++)
					Y[i][j] = Y[j][i] = A[i] == -1 || A[j] == -1 ? 0 : X[A[i]][A[j]];
		}

		return Y;
	}
}

void mat_add(double **X, double **Y, unsigned int n) {
	if (X != NULL && Y != NULL) {
		unsigned int i, j;

		for (i = 0; i < n; i++)
			for (j = 0; j < n; j++)
				X[i][j] += Y[i][j];
	}
}

void mat_sub(double **X, double **Y, unsigned int n) {
	if (X != NULL && Y != NULL) {
		unsigned int i, j;

		for (i = 0; i < n; i++)
			for (j = 0; j < n; j++)
				X[i][j] += Y[i][j];
	}
}

unsigned int mat_nonzeros(double **X, unsigned int n) {
	if (X == NULL) {
		return 0;
	} else {
		unsigned int i, j, nz = 0;
		for (i = 0; i < n; i++)
			for (j = i + 1; j < n; j++)
				nz += X[i][j] != 0;
		return nz;
	}
}

void mat_reciprocal(double **X, unsigned int n) {
	if (X != NULL) {
		unsigned int i, j;

		for (i = 0; i < n; i++)
			for (j = 0; j < n; j++)
				if (X[i][j] != 0)
					X[i][j] = 1 / X[i][j];
	}
}

double **mat_extend(double **X, unsigned int n, unsigned int m) {
	if (X == NULL) {
		return NULL;
	} else {
		double **Y = NULL;

		if (n <= m) {
			unsigned int i, j;
			Y = mat_double_alloc(m, m);
			for (i = 0; i < n; i++)
				for (j = 0; j < n; j++)
					Y[i][j] = X[i][j];
		}

		return Y;
	}
}

double mat_sum(double **X, unsigned int n, unsigned int diag) {
	if (X == NULL) {
		return 0;
	} else {
		double sum = 0;
		int i, j;

		for (i = 0; i < n; i++)
			for (j = 0; j < n; j++)
				if (ABS(i - j) >= diag)
					sum += X[i][j];
		return sum;
	}
}

double mat_sum2(double **X, unsigned int n, unsigned int diag) {
	if (X == NULL) {
		return 0;
	} else {
		double sum = 0;
		int i, j;

		for (i = 0; i < n; i++)
			for (j = 0; j < n; j++)
				if (ABS(i - j) >= diag)
					sum += X[i][j] * X[i][j];
		return sum;
	}
}

double mat_sum_prod(double **X, double **Y, unsigned int n, unsigned int diagX, unsigned int diagY) {
	if (X == NULL || Y == NULL) {
		return 0;
	} else {
		int i, j, k;
		double sum = 0;

		for (i = 0; i < n; i++)
			for (j = 0; j < n; j++)
				for (k = 0; k < n; k++)
					if (ABS(i - k) >= diagX && ABS(j - k) >= diagY)
						sum += X[i][k] * Y[k][j];
		return sum;
	}
}

/******************************
       Matrix index
 ******************************/

double **mat_to_list(double **M, unsigned int map_size, unsigned int *list_size) {
	if (M == NULL) {
		return NULL;
	} else {
		size_t i, j, c;
		double **L = NULL;

		for (i = c = 0; i < map_size; i++)
			for (j = i; j < map_size; j++)
				c += M[i][j] != 0;

		*list_size = c;
		L = mat_double_alloc(c, 3);

		for (i = c = 0; i < map_size; i++)
			for (j = i; j < map_size; j++)
				if (M[i][j] != 0) {
					L[c][0] = i;
					L[c][1] = j;
					L[c][2] = M[i][j];
					c++;
				}
		return L;
	}
}

static double **kronecker_list(double **L, unsigned int map_size, unsigned int *list_size) {
	double **T;
	size_t i, j, c;

	for (i = c = 0; i < *list_size; i++)
		for (j = 0; j < *list_size; j++)
			c += L[i][0] * map_size + L[j][0] <= L[i][1] * map_size + L[j][1];

	T = mat_double_alloc(c, 5);

	for (i = c = 0; i < *list_size; i++)
		for (j = 0; j < *list_size; j++) {
			if (L[i][0] * map_size + L[j][0] <= L[i][1] * map_size + L[j][1]) {
				T[c][0] = L[i][0];
				T[c][1] = L[i][1];
				T[c][2] = L[j][0];
				T[c][3] = L[j][1];
				T[c][4] = L[i][2] * L[j][2];
				c++;
			}
		}
	*list_size = c;
	return T;
}

static int indx_cmp(const void *a, const void *b) {
	struct index *x = (struct index *)a;
	struct index *y = (struct index *)b;
	int i;

	if (x->n != y->n) {
		return x->n - y->n;
	} else {
		for (i = 0; i < 4; i++)
			if (x->val[i] != y->val[i])
				return x->val[i] - y->val[i];
	}

	return 0;
}

static int len_cmp(const void *a, const void *b) {
	int x = *(int *)a;
	int y = *(int *)b;

	return x - y;
}

static int rank(int buf[4]) {
	int i, c = 0, tmp[4];

	qsort(buf, 4, sizeof(buf[0]), len_cmp);

	for (i = 0; i < 4; i++)
		tmp[i] = -1;
	tmp[c] = buf[0];
	for (i = 1; i < 4; i++)
		if (buf[i] != tmp[c])
			tmp[++c] = buf[i];

	for (i = 0; i < 4; i++)
		buf[i] = tmp[i];

	return c + 1;
}

static void map_values(int ind[4], int val[4], int n) {
	int i, j;
	for (i = 0; i < 4; i++) {
		// search position j of ind[i] in val[]
		for (j = 0; j < n && val[j] != ind[i]; j++) ;
		if (j == n) {
			fprintf(stderr, "ERROR!!\n");
			exit(1);
		}
		ind[i] = j;
	}
}

struct index *mat_index(double **M, unsigned int n, unsigned int *indx_len) {
	double **L = mat_to_list(M, n, indx_len);
	double **K = kronecker_list(L, n, indx_len);
	struct index *I;
	int i;

	I = (struct index *)malloc(*indx_len * sizeof(struct index));
	for (i = 0; i < *indx_len; i++) {
		I[i].ind[0] = I[i].val[0] = K[i][0];
		I[i].ind[1] = I[i].val[1] = K[i][1];
		I[i].ind[2] = I[i].val[2] = K[i][2];
		I[i].ind[3] = I[i].val[3] = K[i][3];
		I[i].x = K[i][4];
		I[i].n = rank(I[i].val);
		map_values(I[i].ind, I[i].val, I[i].n);
	}

	qsort(I, *indx_len, sizeof(struct index), indx_cmp);
	mat_double_free(L);
	mat_double_free(K);

	return I;
}

/*****************************
        SUPPORT FUNCTIONS
 *****************************/

static int mat_read_type(char *file) {
	char cmd[strlen(file) + 7];
	FILE *in = ISGZ(file) ? popen(CMD(cmd, file), "r") : fopen(file, "r");
	char buf[7] = { '\0' };

	if (in == NULL || fgets(buf, 7, in) == NULL) {
		return -1;
	} else {
		fclose(in);
		return strcmp(buf, "PFRMAT") == 0 ? CASP_FORMAT : buf[0] == '>' ? FASTA_FORMAT : SIMP_FORMAT;
	}
}

static double **mat_read_simple(char *file, unsigned int *n, unsigned int diag) {
	double x, **M = NULL;
	unsigned int i, j;
	char cmd[strlen(file) + 7];
	FILE *in = ISGZ(file) ? popen(CMD(cmd, file), "r") : fopen(file, "r");

	if (in != NULL && fscanf(in, "%u\n", n) != EOF) {
		M = mat_double_alloc(*n, *n);
		while (fscanf(in, "%u %u %lf\n", &i, &j, &x) != EOF)
			if (i < *n && j < *n && i <= j && ABS(j - i) >= diag) {
				M[i][j] = M[j][i] = x;
				//printf("C1 %u %u %f\n",i,j,x);
			}

		fclose(in);
	}
	return M;
}

static unsigned int mat_read_num_lines(char *file) {
	char *line = NULL;
	size_t len = 0;
	ssize_t read;
	char cmd[strlen(file) + 7];
	FILE *in = ISGZ(file) ? popen(CMD(cmd, file), "r") : fopen(file, "r");
	unsigned int n = 0;

	if (in != NULL) {
		while ((read = getline(&line, &len, in)) != -1 && strlen(line) > 1)
			n++;
		free(line);
		fclose(in);
	}

	return n;
}

static unsigned int get_seq_len(char *buf) {
	unsigned int n = 0;
	while (*buf)
		n += isalpha(*buf++) != 0;
	return n;
}

static double **mat_read_casp(char *file, unsigned int *n, unsigned int diag) {
	double x, **M = NULL, tmp;
	unsigned int i, j;
	char cmd[strlen(file) + 7];
	FILE *in = ISGZ(file) ? popen(CMD(cmd, file), "r") : fopen(file, "r");

	char *line = NULL;
	size_t len = 0;
	ssize_t read;
	int data = 0;

	*n = 0;
	if (in != NULL) {
		unsigned int N = mat_read_num_lines(file), k = 0;

		// Skip the first part of the file
		while (k++ < N - 1 && (read = getline(&line, &len, in)) != -1 && strncmp(line, "MODEL", 5) != 0) ;
		// From here we get the sequence
		while (k++ < N - 1 && (read = getline(&line, &len, in)) != -1 && (data = sscanf(line, "%u %u %lf %lf %lf\n", &i, &j, &tmp, &tmp, &x)) == 0)
			*n += get_seq_len(line);
		// From here we should read the predictions
		if (k < N - 1 && read != -1 && data == 5) {
			M = mat_double_alloc(*n, *n);
			if (i <= *n && j <= *n && i <= j && i >= 1 && j >= 1 && ABS(j - i) >= diag) {
				M[i - 1][j - 1] = M[j - 1][i - 1] = x;
				//printf("C2 %u %u %f\n",i-1,j-1,x);
			}
		} else {	// File contanins some errors: k>=N-1 || read == -1 || data != 5, quit here
			*n = 0;
			return NULL;
		}
		while (k++ < N - 1 && (read = getline(&line, &len, in)) != -1 && (data = sscanf(line, "%u %u %lf %lf %lf\n", &i, &j, &tmp, &tmp, &x)) == 5) {
			if (i <= *n && j <= *n && i <= j && i >= 1 && j >= 1 && ABS(j - i) >= diag) {
				M[i - 1][j - 1] = M[j - 1][i - 1] = x;
				//printf("C2 %u %u %f\n",i-1,j-1,x);
			}
		}
		if (k != N) {	// We should be here because k == N
			mat_double_free(M);
			*n = 0;
		} else {	// Last line should contain END
			getline(&line, &len, in);
			if (strncmp(line, "END", 3) != 0) {
				mat_double_free(M);
				*n = 0;
			}
		}

		fclose(in);
	}

	return M;
}

static int *mat_aln_read_simp(char *file, unsigned int *m, unsigned int *n) {
	FILE *in;

	*m = *n = 0;

	if ((in = fopen(file, "r")) == NULL || fscanf(in, "%u %u\n", m, n) == EOF) {
		return NULL;
	} else {
		unsigned int i, j;
		int *A = (int *)malloc(*m * sizeof(int)), last;

		for (i = 0; i < *m; i++)
			A[i] = -1;

		while (fscanf(in, "%u:%u\n", &i, &j) != EOF) {
			// check bounds and 1-1 match
			if (i >= *m || j >= *n || A[i] != -1) {
				free(A);
				return NULL;
			} else {
				A[i] = j;
			}
		}
		fclose(in);
		// check consistency: i < j && A[i] != -1 && A[j] != -1 => A[i] < A[j]
		for (i = 0, last = -1; i < *m; i++) {
			if (A[i] != -1) {
				if (last == -1 || last < A[i]) {
					last = A[i];
				} else {	// last != -1 && last >= A[i]
					free(A);
					return NULL;
				}
			}
		}
		return A;
	}
}

static int *mat_aln_read_fasta(char *file, unsigned int *m, unsigned int *n, char **fasta1, char **fasta2) {
	FILE *in;
	int *A = NULL;

	*fasta1 = *fasta2 = NULL;

	if ((in = fopen(file, "r")) != NULL) {
		unsigned int i, k1, k2, N, curr_size = MAX_FASTA, check = 0;
		char *line = NULL, *fasta = NULL, *f1 = NULL, *f2 = NULL;
		size_t len = 0;
		ssize_t read;

		while ((read = getline(&line, &len, in)) != -1 && (check == 0 || check == 1 || line[0] != '>')) {
			// Delete trailing newline
			if (line[read - 1] == '\n')
				line[read - 1] = '\0';
			// Skip header
			if (line[0] != '>') {
				if (strlen(fasta) + strlen(line) > curr_size) {
					// Warning: doesn't deal with reallocation errors
					curr_size = 2 * (strlen(fasta) + strlen(line)) + 1;
					fasta = (char *)realloc(fasta, curr_size);
				}
				strncat(fasta, line, strlen(line));
			} else {
				if (check == 1)
					f1 = (char *)realloc(fasta, strlen(fasta) + 1);
				curr_size = MAX_FASTA;
				fasta = (char *)malloc((curr_size + 1) * sizeof(char));
				check++;
			}
		}
		f2 = (char *)realloc(fasta, strlen(fasta) + 1);

		// Length check for aligned sequences
		if (strlen(f1) != strlen(f2)) {
			free(f1);
			free(f2);
			return NULL;
		}
		// Compute alignment
		N = strlen(f1);
		for (i = *m = *n = 0; i < N; i++) {
			if (f1[i] != '-')
				(*m)++;
			if (f2[i] != '-')
				(*n)++;
		}
		A = (int *)malloc(*m * sizeof(int));
		for (i = 0; i < *m; i++)
			A[i] = -1;
		for (i = k1 = k2 = 0; i < N; i++) {
			if (f1[i] != '-' && f2[i] != '-')
				A[k1] = k2;
			if (f1[i] != '-')
				k1++;
			if (f2[i] != '-')
				k2++;
		}

		// Remove gaps from fasta sequences
		for (i = k1 = k2 = 0; i < N; i++) {
			if (f1[i] != '-')
				f1[k1++] = f1[i];
			if (f2[i] != '-')
				f2[k2++] = f2[i];
		}
		f1[*m] = f2[*n] = '\0';
		f1 = (char *)realloc(f1, strlen(f1) + 1);
		f2 = (char *)realloc(f2, strlen(f2) + 1);
		*fasta1 = f1;
		*fasta2 = f2;

		free(line);
		fclose(in);
	}

	return A;
}
