/*********************************************************************
This file is part of the ccpro-1.0 software. 

Copyright (C) 2020  Pietro Di Lena

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
**********************************************************************/

#ifndef MAT_H
#define MAT_H

// Indexing for e2map calculation
struct index {
	int ind[4];		// position indexes: i=val[ind[0]] j=val[ind[1]] k=val[ind[0]] l=val[ind[1]]
	int val[4];		// sorted unique values: val[i]<val[i+1] 0<=i<=n
	int n;			// number of unique values
	double x;		// x=M[i][j]*M[k][l]
};

/******************************
        Matrix allocation
 ******************************/

// double matrix allocation/deallocation

double **mat_double_alloc(unsigned int row, unsigned int col);

long double **mat_ldouble_alloc(unsigned int row, unsigned int col);

void mat_double_free(double **M);

/******************************
        Matrix I/O
 ******************************/

// Read and print matrices

// Read a matrix
// Automatically recognizes CASP and simple format
// if diag > 0 all values on diagonals < diag are set to zero
double **mat_read(char *file, unsigned int *n, unsigned int diag);

// Read an alignment
// Automatically recognized fasta and simple format
// m = map1 size, n = map2 size
int *mat_aln_read(char *file, unsigned int *m, unsigned int *n);

// Save X in simple format into filname
int mat_save(double **X, unsigned int n, char *filename);

// Print X in simple format
void mat_print(double **X, unsigned int n);

// Print X in full format
void mat_full_print(double **X, unsigned int n);

/******************************
       Matrix operations
 ******************************/

// Return 1 iff X's main diagonal is zero
int mat_zero_diag(double **X, unsigned int n);

// Return the trace of symmetric matrix X
double mat_trace(double **X, unsigned int n);

// Return the trace of the squared diagonal elements of symmetric matrix X
double mat_trace2(double **X, unsigned int n);

// Return the trace of the matrix product XY, where X and Y are
// symmetric matrices of size n x n
double mat_trace_prod(double **X, double **Y, unsigned int n);

// Return the trace of the matrix product XY^T, where X and Y are
// matrices of size n x m
double mat_trace_prod_gen(double **X, double **Y, unsigned int m, unsigned int n);

// Return the trace of the matrix product XY^A, where Y^A is the Y matrix
// recoded with respect to alignment A
double mat_trace_prod_aln(double **X, double **Y, int *A, unsigned int n);

// Return the summ of X's entries, ignoring diagonals < diag 
double mat_sum(double **X, unsigned int n, unsigned int diag);

// Return the sum of X's squared entries, ignoring diagonals < diag
double mat_sum2(double **X, unsigned int n, unsigned int diag);

// Return the sum of the XY's entries, ignoring X's diagonals < diagX and Y's diagonals z diagY
double mat_sum_prod(double **X, double **Y, unsigned int n, unsigned int diagX, unsigned int diagY);

// Rounds off to zero all entries in X lower than cutoff
void mat_round(double **X, unsigned int n, double cutoff);

// Returns the Kronecker product between X and Y
double **mat_kronecker_prod(double **X, double **Y, unsigned int n);

// Returns the matrix X recoded with respect to alignment A
double **mat_align(double **X, int *A, unsigned int m);

// Adds Y to X
void mat_add(double **X, double **Y, unsigned int n);

// Subtracts Y to X
void mat_sub(double **X, double **Y, unsigned int n);

// Computes the reciprocal of X, where every non-zero entryx  is recoded as 1/x
void mat_reciprocal(double **X, unsigned int n);

// Extends n x n matrix X to an m x m matrix by addind zero rows/columns
double **mat_extend(double **X, unsigned int n, unsigned int m);

// Returns the number of non-zero entries in the the upper triangualr
// portion of X
unsigned int mat_nonzeros(double **X, unsigned int n);

/******************************
       Matrix indexes
 ******************************/

// Return the list of 0 <= i <= j < n indices and non-zero values x such that M[i][j]=x 
double **mat_to_list(double **M, unsigned int n, unsigned int *list_size);

// Return the index structure for the Kronecked product of the list returned by mat_to_list 
struct index *mat_index(double **M, unsigned int n, unsigned int *indx_len);

#endif
