/*********************************************************************
This file is part of the ccpro-1.0 software. 

Copyright (C) 2020  Pietro Di Lena

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <time.h>

static unsigned int DIAG = 0;	// Remove diagonal elements < DIAG
static double PROB = 1;
static double MINV = 0;
static double MAXV = 1;
static char *SIZE = NULL;

static int read_ui(char *num, unsigned int *n) {
	char *ptr;
	int base = 10;

	*n = strtoul(num, &ptr, base);
	return num == ptr || errno != 0 || (errno == 0 && num && *ptr != 0);
}

static int read_d(char *num, double *n) {
	char *ptr;

	*n = strtod(num, &ptr);
	return num == ptr || errno != 0 || (errno == 0 && num && *ptr != 0);
}

void usage(void) {
	fprintf(stderr, "Usage: randmap [options] <size>\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "\t-d, --diag=<n>       Set to 0 diagonals lower than <n>. Default: %u\n", DIAG);
	fprintf(stderr, "\t-p, --prob=<p>       Probability of sampling an element. Default: %.2f\n", PROB);
	fprintf(stderr, "\t-m, --min=<m>        Min value. Default: %f\n", MINV);
	fprintf(stderr, "\t-M, --max=<m>        Max value. Default: %f\n", MAXV);
}

int parse_args(int argc, char *argv[]) {
	int opt = 0, long_index = 0;

	static struct option long_options[] = {
		{"diag", required_argument, 0, 'd'},
		{"prob", required_argument, NULL, 'p'},
		{"min", required_argument, 0, 'm'},
		{"max", required_argument, 0, 'M'},
		{"help", no_argument, 0, 'h'},
		{0, 0, 0, 0}
	};

	while ((opt = getopt_long(argc, argv, "d:p:m:M:", long_options, &long_index)) != -1) {
		switch (opt) {
		case 'd':
			if (read_ui(optarg, &DIAG) != 0) {
				fprintf(stderr, "Error reading -d argument: %s\n", optarg);
				return 1;
			}
			break;
		case 'p':
			if (read_d(optarg, &PROB) != 0) {
				fprintf(stderr, "Error reading -p argument: %s\n", optarg);
				return 1;
			}
			if (PROB < 0 || PROB > 1) {
				fprintf(stderr, "Error: probability should be a value between 0 and 1\n");
				return 1;
			}
			break;
		case 'm':
			if (read_d(optarg, &MINV) != 0) {
				fprintf(stderr, "Error reading -m argument: %s\n", optarg);
				return 1;
			}
			break;
		case 'M':
			if (read_d(optarg, &MAXV) != 0) {
				fprintf(stderr, "Error reading -M argument: %s\n", optarg);
				return 1;
			}
			break;
		case 'h':
			usage();
			return 0;
		default:
			fprintf(stderr, "Unrecognized option -%c\n\n", opt);
			return 1;
		}
	}

	if (argc - optind < 1) {
		usage();
		return 1;
	} else if (argc - optind > 1) {
		int i;
		fprintf(stderr, "Unrecognized arguments:");
		for (i = optind + 1; i < argc; i++)
			fprintf(stderr, " %s", argv[i]);
		fprintf(stderr, "\n\n");
		return 1;
	}

	SIZE = argv[optind];

	return 0;
}

int main(int argc, char *argv[]) {
	unsigned int n, i, j;

	srandom(time(NULL));

	if (parse_args(argc, argv) != 0)
		return 1;

	if (read_ui(SIZE, &n) != 0) {
		fprintf(stderr, "Error converting to unsigned int %s\n", SIZE);
		return 1;
	}

	printf("%d\n", n);
	for (i = 0; i < n; i++)
		for (j = i + DIAG; j < n; j++)
			if (1.0 * random() / RAND_MAX < PROB)
				printf("%d %d %.4f\n", i, j, (1.0 * random() / RAND_MAX) * (MAXV - MINV) + MINV);

	return 0;
}
